# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import contrast
from contrast.agent import scope as scope_
from contrast.agent.policy.trigger_node import TriggerNode
from contrast.agent.middlewares.base_middleware import BaseMiddleware
from contrast.agent.asgi import (
    track_scope_sources,
    ASGIRequest,
    ASGIResponse,
)
from contrast.utils.decorators import log_time_cm, cached_property

from contrast.utils import Profiler
from contrast.extern import structlog as logging


logger = logging.getLogger("contrast")

DEFAULT_ASGI_NAME = "asgi_app"


class ASGIMiddleware(BaseMiddleware):
    def __init__(self, asgi_app, app_name=None):
        self.app_name = (
            app_name
            if app_name is not None
            else getattr(asgi_app, "__name__", DEFAULT_ASGI_NAME)
        )

        super().__init__()

        self.asgi_app = asgi_app

    async def __call__(self, scope, receive, send) -> None:
        if scope.get("type") == "websocket":
            await self.call_without_agent_async(scope, receive, send)
        else:
            self.request_path = scope.get("path", "")
            with Profiler(self.request_path):
                request = ASGIRequest(scope, receive)
                environ = await request.to_wsgi_environ()

                context = self.should_analyze_request(environ)
                if context:
                    with contrast.CS__CONTEXT_TRACKER.lifespan(context):
                        await self.call_with_agent(context, request, send)
                        return

                await self.call_without_agent_async(scope, receive, send)

    async def call_with_agent(self, context, request, send) -> None:
        self.log_start_request_analysis()
        track_scope_sources(context, request.scope)

        try:
            self.prefilter()

            response = ASGIResponse(send)
            with log_time_cm("app code and get response"):
                await self.asgi_app(
                    request.scope, request.fake_receive, response.fake_send
                )

            context.extract_response_to_context(response)

            self.postfilter(context)
            self.check_for_blocked(context)
            self.handle_ensure(context, context.request)

            await response.call_send()

        except Exception as e:
            self.handle_ensure(context, context.request)
            # TODO: PYT-2127
            _ = self.handle_exception(e)
            return
        finally:
            self.log_end_request_analysis()
            if self.settings.is_assess_enabled():
                contrast.STRING_TRACKER.ageoff()

    async def call_without_agent_async(self, scope, receive, send) -> None:
        super().call_without_agent()
        with scope_.contrast_scope():
            await self.asgi_app(scope, receive, send)

    @cached_property
    def trigger_node(self):
        """
        trigger node used by reflected xss postfilter rule
        """
        method_name = self.app_name

        module, class_name, args, instance_method = self._process_trigger_handler(
            self.asgi_app
        )

        return (
            TriggerNode(module, class_name, instance_method, method_name, "RETURN"),
            args,
        )

    @cached_property
    def name(self):
        return "asgi"



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.

import sys
import platform
import atexit

from typing import Optional

import contrast
from contrast import __version__, AGENT_CURR_WORKING_DIR
from contrast.agent import agent_lib, thread_watcher, patch_controller
from contrast.agent.assess.rules.providers.enable import enable_providers
from contrast.agent.settings import Settings
from contrast.reporting import ReportingClient, teamserver_messages
from contrast.reporting.reporting_client import MAX_ATTEMPTS
from contrast.assess_extensions import cs_str
from contrast.utils.loggers.logger import (
    setup_basic_agent_logger,
    setup_agent_logger,
    setup_security_logger,
)
from contrast.utils import timer
from contrast.utils.library_reader.library_reader import LibraryReader
from contrast.utils.namespace import Namespace
from contrast.utils.decorators import fail_loudly

logger = setup_basic_agent_logger()
LOGS_SEPARATOR = "-" * 120

if not contrast.telemetry_disabled():
    from contrast.agent.telemetry import Telemetry
else:
    Telemetry = None


def stop_telemetry(telemetry):
    if telemetry.message_q:
        if not telemetry.message_q.empty():
            telemetry.send_messages()

        assert telemetry.message_q.empty()


class module(Namespace):
    is_initialized = False
    id: Optional[int] = None
    settings: Optional[Settings] = None
    reporting_client: Optional[ReportingClient] = None
    library_reader: Optional[LibraryReader] = None


def _log_environment():
    """
    Log current working directory, python version and pip version
    """
    banner = f"{'-' * 50}ENVIRONMENT{'-' * 50}"
    logger.debug(banner)
    logger.debug("Current Working Dir: %s", AGENT_CURR_WORKING_DIR)
    logger.debug("Python Version: %s", sys.version)
    logger.debug("Framework Version: %s", module.settings.framework)
    logger.debug("Server Version: %s", module.settings.server)
    logger.debug("Contrast Python Agent Version: %s", __version__)
    logger.debug("Executable: %s", sys.executable)

    try:
        platform_str = platform.platform()
    except Exception:
        try:
            platform_str = platform.platform(terse=True)
        except Exception:
            platform_str = "unknown"

    logger.debug("Platform %s", platform_str)

    try:
        import pip

        logger.debug("Pip Version: %s", pip.__version__)
    except Exception:
        logger.debug("Pip not found")

    logger.debug(banner)


def _warn_for_misleading_config():
    protect_enabled = module.settings.is_protect_enabled()
    assess_enabled = module.settings.is_assess_enabled()

    logger.info("Protect: %s", protect_enabled)
    logger.info("Assess: %s", assess_enabled)

    if protect_enabled and module.settings.config.get("assess.enable"):
        logger.warning("Protect is running but Assess is enabled in local config")
        logger.warning("Defaulting to Protect behavior only")

    if not protect_enabled and module.settings.config.get("protect.enable", True):
        logger.warning("Protect enabled in local config but disabled by Teamserver")

    if (
        module.settings.is_agent_config_enabled()
        and not protect_enabled
        and not assess_enabled
    ):
        logger.warning("Neither Protect nor Assess is running")


def is_initialized() -> bool:
    return module.is_initialized


def successful_startup_msgs():
    if not _send_startup_msg_with_retries(teamserver_messages.ServerStartup()):
        return False
    if not _send_startup_msg_with_retries(teamserver_messages.ApplicationStartup()):
        return False
    return True


def _send_startup_msg_with_retries(msg):
    attempt = 1
    while True:
        if _send_startup_msg(msg):
            return True

        attempt += 1
        if attempt > MAX_ATTEMPTS:
            break

        logger.debug("%s: will retry - sleeping for 1 second", msg.class_name)
        timer.sleep(1)

    # TODO: PYT-2612 use the commented-out logic here instead
    # logger.error(
    #     "%s: Unexpected error from Contrast UI. Will not initialize Contrast Agent."
    # )
    # return False
    return True


def _send_startup_msg(msg):
    """
    Send a single startup message (ServerStartup or ApplicationStartup) immediately.
    This function runs in the main thread, and is blocking.

    Return False if the message needs to be retried, True otherwise.
    Note that we only need to retry if we get an ambiguous response (5xx).
    """
    response = module.reporting_client.send_message(msg)
    if response is None:
        return False
    msg.process_response(response, module.reporting_client)
    return response.status_code < 500


def initialize_libraries():
    """
    If enabled, read libraries from the application
    :return: True
    """
    if not module.settings.is_analyze_libs_enabled():
        return

    # Passing callback to send message due to a circular import issue, and a
    # deadlock occurring if something is imported inside of a function running in
    # this thread
    module.library_reader = LibraryReader(
        module.settings,
        send_ts_message_func=module.reporting_client.add_message,
    )
    module.library_reader.start_library_analysis_thread()


@fail_loudly("Unable to initialize Contrast Agent Settings.", return_value=False)
def initialize_settings(runner: bool = False):
    """
    Initialize agent settings.

    Returns True on settings being initialized and False if any failure.
    """
    module.settings = Settings(runner=runner)
    return True


def get_settings():
    return module.settings


def get_reporting_client():
    return module.reporting_client


def initialize(runner: bool = False, name: Optional[str] = None):
    """
    If this method is called more than once per process, we use the is_initialized
    flag to only run the following work once:
        - library analysis thread initialization
        - turning on patches
        - hardcoded rule providers
        - scanning config rules
        - environment logging
        - loading common configuration
        - initializing settings

    In order to avoid a warning message, callers should check is_initialized()
    prior to calling this function.
    """
    # TODO: PYT-2620 Implement lock for agent state initialization
    if module.is_initialized:
        logger.warning("Attempted to initialize agent state more than once")
        return

    module.id = id(module)
    name = name or __name__

    logger.info('Initializing Contrast Agent "%s" [id=%s]', name, module.id)
    logger.info("Contrast Python Agent Version: %s\n", __version__)

    if not initialize_settings(runner=runner):
        return

    if module.settings is None or not module.settings.is_agent_config_enabled():
        logger.warning("Contrast Agent is not enabled.")
        return

    setup_agent_logger(module.settings.config)
    setup_security_logger(module.settings.config)
    module.settings.config.log_config()

    _log_environment()
    _warn_for_misleading_config()

    cs_str.init_contrast_scope_cvars()

    module.reporting_client = ReportingClient()
    module.reporting_client.start()

    if not successful_startup_msgs():
        logger.error("Unable to initialize Contrast Agent.", direct_to_teamserver=1)
        return

    # This MUST happen after the initialization calls for TeamServer messaging
    # (sending server start and application start) to ensure that TeamServer will
    # accept the messages sent by our background reporting threads
    thread_watcher.ensure_running(module)

    initialize_libraries()
    patch_controller.enable_patches()

    if module.settings.is_assess_enabled():
        # For now agent runtime starts before config scanning
        # this will be reset when time_limit_threshold is reached,
        # it doesn't symbolize the total agent runtime for all time.
        module.settings.agent_runtime_window = timer.now_ms()

        enable_providers()
        # NOTE: This should stay in middleware
        # scan_configs_in_thread()

    if module.settings.is_protect_enabled() and not agent_lib.initialize():
        logger.error("Fatal: Unable to initialize agent-lib. Disabling Contrast.")
        module.settings.config.put("enable", False)
        return

    logger.info(
        "Finished Initializing Contrast Agent %s [id=%s] \n\n%s\n",
        name,
        module.id,
        LOGS_SEPARATOR,
    )

    if Telemetry is not None:
        contrast.TELEMETRY = Telemetry()
        contrast.TELEMETRY.start()
        atexit.register(stop_telemetry, contrast.TELEMETRY)

    module.is_initialized = True



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.middlewares.route_coverage.pyramid_routes import PyramidRoutesMixin
from contrast.wsgi.middleware import WSGIMiddleware
from contrast.agent.middlewares.app_finder import get_original_app_or_fail
from contrast.extern import structlog as logging
from contrast.utils.decorators import fail_quietly, cached_property

from pyramid.registry import Registry

logger = logging.getLogger("contrast")


class PyramidWSGIMiddleware(PyramidRoutesMixin, WSGIMiddleware):
    """
    A subclass of the WSGI middleware that provides pyramid route coverage.
    "configuration scanning" in pyramid is handled dynamically with triggers.

    This is not a pyramid-style tween - it must wrap pyramid's WSGI application
    directly. The WSGI app is typically returned by a call to some Configurator
    instance's `make_wsgi_app()`.
    """

    def __init__(self, wsgi_app, orig_pyramid_registry=None):
        """
        This deviates slightly from typical WSGI Middleware API, because we need
        `registry` to get framework-specific information about the application.

        The application's `Registry` is commonly available as an attribute on the
        `Configurator` instance used to construct the application. It is also available
        as an attribute of the original WSGI application object returned by
        `Configurator.make_wsgi_app()`. Note that we can't guarantee that the WSGI app
        passed to this constructor has a `registry` attribute, since the original
        Pyramid WSGI app might have already been wrapped by other WSGI middlewares.

        @param wsgi_app: a WSGI application object for the Pyramid application to be
            instrumented by Contrast
        @param registry: (optional) the `pyramid.registry.Registry` instance
            corresponding to the wsgi_app - if not provided, we make a best effort to
            find it on the provided app
        """
        self.registry = (
            orig_pyramid_registry
            if orig_pyramid_registry is not None
            and isinstance(orig_pyramid_registry, Registry)
            else get_original_app_or_fail(wsgi_app, Registry)
        )

        app_name = self._get_app_name()
        super().__init__(wsgi_app, app_name)

    @fail_quietly(
        "Unable to get Pyramid application name", return_value="Pyramid Application"
    )
    def _get_app_name(self):
        return self.registry.package_name

    @cached_property
    def name(self):
        return "pyramid"



# SPDX-License-Identifier: MIT OR Apache-2.0
# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the MIT License.  See the LICENSE file in the root of this
# repository for complete details.

"""
Logger wrapper and helper class.
"""

import copy
import sys
import threading

from pickle import PicklingError
from typing import IO, Any, BinaryIO, Dict, Optional, TextIO

from contrast.extern.structlog._utils import until_not_interrupted


WRITE_LOCKS: Dict[IO[Any], threading.Lock] = {}


def _get_lock_for_file(file: IO[Any]) -> threading.Lock:
    global WRITE_LOCKS

    lock = WRITE_LOCKS.get(file)
    if lock is None:
        lock = threading.Lock()
        WRITE_LOCKS[file] = lock

    return lock


class PrintLogger:
    """
    Print events into a file.

    :param file: File to print to. (default: `sys.stdout`)

    >>> from contrast.extern.structlog import PrintLogger
    >>> PrintLogger().msg("hello")
    hello

    Useful if you follow
    `current logging best practices <logging-best-practices>`.

    Also very useful for testing and examples since logging is finicky in
    doctests.
    """

    def __init__(self, file: Optional[TextIO] = None):
        self._file = file or sys.stdout
        self._write = self._file.write
        self._flush = self._file.flush

        self._lock = _get_lock_for_file(self._file)

    def __getstate__(self) -> str:
        """
        Our __getattr__ magic makes this necessary.
        """
        if self._file is sys.stdout:
            return "stdout"

        elif self._file is sys.stderr:
            return "stderr"

        raise PicklingError(
            "Only PrintLoggers to sys.stdout and sys.stderr can be pickled."
        )

    def __setstate__(self, state: Any) -> None:
        """
        Our __getattr__ magic makes this necessary.
        """
        if state == "stdout":
            self._file = sys.stdout
        else:
            self._file = sys.stderr

        self._write = self._file.write
        self._flush = self._file.flush
        self._lock = _get_lock_for_file(self._file)

    def __deepcopy__(self, memodict: Dict[Any, Any] = {}) -> "PrintLogger":
        """
        Create a new PrintLogger with the same attributes. Similar to pickling.
        """
        if self._file not in (sys.stdout, sys.stderr):
            raise copy.error(
                "Only PrintLoggers to sys.stdout and sys.stderr "
                "can be deepcopied."
            )

        newself = self.__class__(self._file)

        newself._write = newself._file.write
        newself._flush = newself._file.flush
        newself._lock = _get_lock_for_file(newself._file)

        return newself

    def __repr__(self) -> str:
        return f"<PrintLogger(file={self._file!r})>"

    def msg(self, message: str) -> None:
        """
        Print *message*.
        """
        with self._lock:
            until_not_interrupted(self._write, message + "\n")
            until_not_interrupted(self._flush)

    log = debug = info = warn = warning = msg
    fatal = failure = err = error = critical = exception = msg


class PrintLoggerFactory:
    r"""
    Produce `PrintLogger`\ s.

    To be used with `structlog.configure`\ 's ``logger_factory``.

    :param file: File to print to. (default: `sys.stdout`)

    Positional arguments are silently ignored.

    .. versionadded:: 0.4.0
    """

    def __init__(self, file: Optional[TextIO] = None):
        self._file = file

    def __call__(self, *args: Any) -> PrintLogger:
        return PrintLogger(self._file)


class BytesLogger:
    r"""
    Writes bytes into a file.

    :param file: File to print to. (default: `sys.stdout`\ ``.buffer``)

    Useful if you follow
    `current logging best practices <logging-best-practices>` together with
    a formatter that returns bytes (e.g. `orjson
    <https://github.com/ijl/orjson>`_).

    .. versionadded:: 20.2.0
    """
    __slots__ = ("_file", "_write", "_flush", "_lock")

    def __init__(self, file: Optional[BinaryIO] = None):
        self._file = file or sys.stdout.buffer
        self._write = self._file.write
        self._flush = self._file.flush

        self._lock = _get_lock_for_file(self._file)

    def __getstate__(self) -> str:
        """
        Our __getattr__ magic makes this necessary.
        """
        if self._file is sys.stdout.buffer:
            return "stdout"

        elif self._file is sys.stderr.buffer:
            return "stderr"

        raise PicklingError(
            "Only BytesLoggers to sys.stdout and sys.stderr can be pickled."
        )

    def __setstate__(self, state: Any) -> None:
        """
        Our __getattr__ magic makes this necessary.
        """
        if state == "stdout":
            self._file = sys.stdout.buffer
        else:
            self._file = sys.stderr.buffer

        self._write = self._file.write
        self._flush = self._file.flush
        self._lock = _get_lock_for_file(self._file)

    def __deepcopy__(self, memodict: Dict[Any, Any] = {}) -> "BytesLogger":
        """
        Create a new BytesLogger with the same attributes. Similar to pickling.
        """
        if self._file not in (sys.stdout.buffer, sys.stderr.buffer):
            raise copy.error(
                "Only BytesLoggers to sys.stdout and sys.stderr "
                "can be deepcopied."
            )

        newself = self.__class__(self._file)

        newself._write = newself._file.write
        newself._flush = newself._file.flush
        newself._lock = _get_lock_for_file(newself._file)

        return newself

    def __repr__(self) -> str:
        return f"<BytesLogger(file={self._file!r})>"

    def msg(self, message: bytes) -> None:
        """
        Write *message*.
        """
        with self._lock:
            until_not_interrupted(self._write, message + b"\n")
            until_not_interrupted(self._flush)

    log = debug = info = warn = warning = msg
    fatal = failure = err = error = critical = exception = msg


class BytesLoggerFactory:
    r"""
    Produce `BytesLogger`\ s.

    To be used with `structlog.configure`\ 's ``logger_factory``.

    :param file: File to print to. (default: `sys.stdout`\ ``.buffer``)

    Positional arguments are silently ignored.

    .. versionadded:: 20.2.0
    """
    __slots__ = ("_file",)

    def __init__(self, file: Optional[BinaryIO] = None):
        self._file = file

    def __call__(self, *args: Any) -> BytesLogger:
        return BytesLogger(self._file)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.asgi.middleware import ASGIMiddleware
from contrast.agent.middlewares.route_coverage.django_routes import DjangoRoutesMixin
from contrast.utils.decorators import fail_quietly

from contrast.agent.assess.rules.config import (
    DjangoHttpOnlyRule,
    DjangoSecureFlagRule,
    DjangoSessionAgeRule,
)

from contrast.extern import structlog as logging
from contrast.utils.decorators import cached_property

# pylint: disable=dangerous-framework-import
from django.http import HttpResponse
from contrast.utils.exceptions.security_exception import SecurityException

logger = logging.getLogger("contrast")


class DjangoASGIMiddleware(DjangoRoutesMixin, ASGIMiddleware):
    def __init__(self, asgi_app):
        self.app_name = self.get_app_name()

        self.config_rules = [
            DjangoSessionAgeRule(),
            DjangoSecureFlagRule(),
            DjangoHttpOnlyRule(),
        ]

        super().__init__(asgi_app, self.app_name)

    @fail_quietly(
        "Unable to get Django application name", return_value="Django Application"
    )
    def get_app_name(self):
        from django.conf import settings

        asgi_application = settings.ASGI_APPLICATION

        return asgi_application.split(".")[0]

    @cached_property
    def name(self):
        return "django"

    def generate_security_exception_response(self):
        return HttpResponse(
            self.OVERRIDE_MESSAGE,
            status=SecurityException.STATUS_CODE,
            content_type="text/html",
        )



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.

from argparse import ArgumentParser
import os
import sys
import warnings

# We'll need to extern this function or investigate replacing it with shutil.which()
# "distutils is deprecated with removal planned for Python 3.12. See the What’s
# New entry for more information."
with warnings.catch_warnings():
    warnings.filterwarnings(
        "ignore",
        "The distutils package is deprecated and slated for removal in Python 3.12. Use setuptools or check PEP 632 for potential alternatives",
        DeprecationWarning,
    )

    from distutils.spawn import find_executable  # pylint: disable=deprecated-module

from contrast import __file__
from contrast.agent.runner import REWRITE_FOR_PYTEST


def runner() -> None:
    parser = ArgumentParser()
    parser.add_argument("--rewrite-for-pytest", action="store_true")
    parser.add_argument("args", nargs="+")

    parsed = parser.parse_args()

    loader_path = os.path.join(os.path.dirname(__file__), "loader")
    os.environ["PYTHONPATH"] = os.path.pathsep.join([loader_path] + sys.path)
    if parsed.rewrite_for_pytest:
        os.environ[REWRITE_FOR_PYTEST] = "true"

    os.execl(find_executable(parsed.args[0]), *parsed.args)



from __future__ import print_function
import os
from pip._vendor.pkg_resources import get_distribution
import sys
from contrast.agent.assess.string_tracker import StringTracker
from contrast.utils.context_tracker import ContextTracker

from contrast.assess_extensions import cs_str

__version__ = get_distribution("contrast-agent").version

CS__CONTEXT_TRACKER = ContextTracker()

STRING_TRACKER = StringTracker()

# PERF: These values are constant for the lifetime of the agent,
# so we compute them only once instead of potentially computing
# them hundreds of times.
AGENT_CURR_WORKING_DIR = os.getcwd()
SORTED_SYS_PATH = sorted(sys.path, key=len, reverse=True)

# --- import aliases ---

from contrast.agent.assess.utils import get_properties  # noqa

TELEMETRY = None


def telemetry_disabled():
    OPT_OUT = os.environ.get("CONTRAST_AGENT_TELEMETRY_OPTOUT", False)
    return OPT_OUT and OPT_OUT.lower() in ("1", "true")



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.extern import structlog as logging

logger = logging.getLogger("contrast")

CONTRAST_ASSESS_METHOD_START = "cs__assess_"


def build_method_name(method_name):
    """
    Builds a name based on the method name

    Example:
        cs__assess_append
    """
    return CONTRAST_ASSESS_METHOD_START + method_name



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import sys
from contrast.extern.wrapt import register_post_import_hook

from contrast.agent.policy.applicator import apply_module_patches
from contrast.agent.policy import patch_manager
from contrast.utils.patch_utils import build_and_apply_patch

MODULE_NAME = "bottle"


def build_prepare_patch(orig_func, patch_policy):
    def prepare_patch(*args, **kwargs):
        """
        Patch for bottle.SimpleTemplate.prepare.

        This is needed because of the unfortunate way bottle calls on
        `html_escape` as a kwarg in the prepare definition in SimpleTemplate.
        See https://github.com/bottlepy/bottle/blob/master/bottle.py#L3952

        Because of this behavior, the `html_escape` func is not our patched
        `html_escape` defined in policy.
        By patching prepare, we intercept its call and instead of allowing
        it to use the default kwarg for `html_escape`, we pass our own
        patched `html_escape` in order to prevent false positive XSS findings.
        """
        # import here instead of top of module because it ensures bottle module exists.
        import bottle

        # html_escape MUST already be patched by policy in order
        # to pass in the patched func to prepare
        kwargs.setdefault("escape_func", bottle.html_escape)
        return orig_func(*args, **kwargs)

    return prepare_patch


def patch_bottle(bottle_module):
    # We ask policy to go ahead and do all bottle patches here (even though policy
    # patches will happen later on) because we MUST have some bottle policy patches
    # already applied for these non-policy patches to work.
    # This would not be necessary if in _enable_patches policy_patches were applied
    # first.
    apply_module_patches(bottle_module)

    build_and_apply_patch(bottle_module.SimpleTemplate, "prepare", build_prepare_patch)


def register_patches():
    register_post_import_hook(patch_bottle, MODULE_NAME)


def reverse_patches():
    bottle_module = sys.modules.get(MODULE_NAME)
    if not bottle_module:
        return

    patch_manager.reverse_patches_by_owner(bottle_module)
    patch_manager.reverse_patches_by_owner(bottle_module.SimpleTemplate)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from pip._vendor.pkg_resources import Requirement

from contrast.agent.policy import constants
from contrast.agent.assess.assess_exceptions import ContrastAssessException
from contrast.agent.policy.policy_node import PolicyNode
from contrast.agent.settings import Settings
from contrast.api.trace_event import TraceEvent
from contrast.utils.string_utils import equals_ignore_case


class SourceNode(PolicyNode):
    ALL = "ALL"

    SOURCE_TAG = "UNTRUSTED"

    def __init__(
        self,
        module,
        class_name,
        instance_method,
        method_name,
        target,
        node_type,
        package,
        tags=None,
        source_name=None,
        version=ALL,
        policy_patch=True,
    ):
        super().__init__(
            module,
            class_name,
            instance_method,
            method_name,
            source_name,
            target,
            tags,
            policy_patch=policy_patch,
        )

        self.package = package
        self.version = version
        self.type = node_type

        if tags is None:
            self.tags = {self.SOURCE_TAG}
        else:
            self.tags.add(self.SOURCE_TAG)

        self.validate()

    @classmethod
    def dynamic(cls, module, class_name, column_name, tags, policy_patch=True):
        """
        Create a dynamic source node, one with the dynamic_source_id property.
        """
        instance_method = True
        method_name = column_name
        target = "RETURN"
        file_name = "dynamic_source"

        node = cls(
            module,
            class_name,
            instance_method,
            method_name,
            target,
            constants.DB_SOURCE_TYPE,
            file_name,
            tags=tags,
            policy_patch=policy_patch,
        )

        node.add_property("dynamic_source_id", node.name)

        return node

    @property
    def node_type(self):
        """
        This is confusing. Sources are Creation action but Propagation type. Oh
        and also Type refers to input type, like parameter, so we have to call
        this node_type. :-/
        """
        return TraceEvent.EventType.PROPAGATION

    def should_apply(self):
        if equals_ignore_case(self.version, self.ALL) or not self.package:
            return True

        settings = Settings()

        if not settings:
            return False

        libraries = [
            lib
            for lib in settings.libraries
            if lib.file_path.lower() == self.package.lower()
        ]

        if not libraries:
            return False

        applicable_library = libraries[0]

        return self._compare_requirement(
            self.package, applicable_library.version, self.version
        )

    @staticmethod
    def _compare_requirement(package, library_version, required_version):
        requirement = SourceNode._convert_to_requirement(package, required_version)

        return requirement and library_version in requirement

    @staticmethod
    def _convert_to_requirement(package, version):
        return Requirement.parse(package + version)

    def validate(self):

        super().validate()

        if not self.targets:
            raise ContrastAssessException(
                f"Source {self.name} did not have a proper target. Unable to create."
            )

        if not self.type:
            raise ContrastAssessException(
                f"Source {self.name} did not have a proper type. Unable to create."
            )

    @classmethod
    def from_dict(cls, framework, obj):
        return cls(
            obj[constants.JSON_MODULE],
            obj.get(constants.JSON_CLASS_NAME, ""),
            obj.get(constants.JSON_INSTANCE_METHOD, True),
            obj[constants.JSON_METHOD_NAME],
            obj[constants.JSON_TARGET],
            obj[constants.JSON_TYPE],
            framework,
            obj.get(constants.JSON_TAGS, []),
            obj.get(constants.JSON_SOURCE_NAME, None),
            obj.get(constants.JSON_VERSION, SourceNode.ALL),
            policy_patch=obj.get(constants.JSON_POLICY_PATCH, True),
        )



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.extern.wrapt import register_post_import_hook

from contrast.agent.policy import patch_manager
from contrast.utils.patch_utils import build_and_apply_patch
from contrast.utils.decorators import fail_loudly


WERKZEUG_SERVING_NAME = "werkzeug.serving"


@fail_loudly("Failed to wrap flask middleware")
def wrap_middleware_args(middleware_cls, args):
    """
    Extract application from arguments and wrap with middleware

    Decorator ensures that None will be returned if any problem occurs
    """
    app = args[2]
    if not isinstance(app, middleware_cls):
        app = middleware_cls(app)
        args = args[:2] + (app,) + args[3:]
    return args


def build_make_server_patch(orig_func, patch_policy):
    # Avoids circular import
    from contrast.flask.middleware import FlaskMiddleware

    del patch_policy  # unused

    def make_server(*args, **kwargs):
        args = wrap_middleware_args(FlaskMiddleware, args) or args
        return orig_func(*args, **kwargs)

    return make_server


def patch_werkzeug_serving(module):
    build_and_apply_patch(module, "make_server", build_make_server_patch)


def register_patches():
    register_post_import_hook(patch_werkzeug_serving, WERKZEUG_SERVING_NAME)


def reverse_patches():
    patch_manager.reverse_module_patches_by_name(WERKZEUG_SERVING_NAME)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import contrast
from contrast.agent import scope as scope_
from contrast.agent.policy.trigger_node import TriggerNode
from fastapi import FastAPI
from starlette.middleware.base import BaseHTTPMiddleware
from starlette.types import ASGIApp
from starlette.requests import Request
from starlette.responses import PlainTextResponse
from contrast.agent.middlewares.base_middleware import BaseMiddleware
from contrast.agent.middlewares.app_finder import get_original_app_or_fail
from contrast.agent.middlewares.response_wrappers.fastapi_response_wrapper import (
    FastApiResponseWrapper,
)
from contrast.agent.middlewares.route_coverage.fastapi_routes import (
    create_fastapi_routes,
)
from contrast.agent.middlewares.route_coverage.common import build_route
from contrast.agent.asgi import starlette_request_to_environ, track_scope_sources
from contrast.extern import structlog as logging
from contrast.utils.exceptions.security_exception import SecurityException
from contrast.utils.decorators import cached_property, fail_quietly, log_time_cm

from contrast.utils import Profiler

logger = logging.getLogger("contrast")


class FastApiMiddleware(BaseMiddleware, BaseHTTPMiddleware):
    def __init__(self, app: ASGIApp, original_app: FastAPI = None) -> None:
        self.app = app
        self.original_app = (
            original_app
            if original_app is not None and isinstance(original_app, FastAPI)
            else get_original_app_or_fail(app, FastAPI)
        )
        self.app_name = self.original_app.title

        self.dispatch_func = self.agent_dispatch_func

        super().__init__()

    async def agent_dispatch_func(self, request: Request, call_next):
        """
        Instead of implementing a __call__ function, we rely on
        BaseHTTPMiddleware's __call__ func which calls self.dispatch_func.
        """
        self.request_path = request.scope.get("path", "")

        with Profiler(self.request_path):
            with scope_.contrast_scope():
                environ = await starlette_request_to_environ(request)

            context = self.should_analyze_request(environ)
            if context:
                with contrast.CS__CONTEXT_TRACKER.lifespan(context):
                    return await self.call_with_agent(context, request, call_next)

            return await self.call_without_agent_async(request, call_next)

    async def call_without_agent_async(self, request: Request, call_next):
        super().call_without_agent()
        with scope_.contrast_scope():
            return await call_next(request)

    async def call_with_agent(self, context, request: Request, call_next):
        scope = request.scope
        self.log_start_request_analysis()

        track_scope_sources(context, scope)
        try:
            self.prefilter()

            with log_time_cm("app code and get response"):
                response = await call_next(request)

            with scope_.contrast_scope():
                wrapped_response = FastApiResponseWrapper(response)

            await self.extract_response_to_context_async(wrapped_response, context)

            self.postfilter(context)
            self.check_for_blocked(context)
            self.handle_ensure(context, context.request)
            return response

        except Exception as e:
            self.handle_ensure(context, context.request)
            response = self.handle_exception(e)
            return response
        finally:
            self.log_end_request_analysis()
            if self.settings.is_assess_enabled():
                contrast.STRING_TRACKER.ageoff()

    @fail_quietly("Unable to get route coverage", return_value={})
    def get_route_coverage(self):
        return create_fastapi_routes(self.original_app)

    @fail_quietly("Unable to get FastAPI view func")
    def get_view_func(self, request):
        if not self.request_path:
            return None

        # Checking for `==` works here because fastapi correctly re-routes
        # if user adds (or doesn't) a /. So this method may be called
        # multiple times per request.
        matching_routes = [
            x for x in self.original_app.routes if x.path == self.request_path
        ]

        if not matching_routes:
            return None

        view_func = matching_routes[0].endpoint
        return view_func

    @fail_quietly("Unable to build route", return_value="")
    def build_route(self, view_func, url):
        return build_route(url, view_func)

    def generate_security_exception_response(self):
        return PlainTextResponse(self.OVERRIDE_MESSAGE, SecurityException.STATUS_CODE)

    @cached_property
    def trigger_node(self):
        """
        FastAPI-specific trigger node used by reflected xss postfilter rule
        """
        method_name = self.app_name

        module, class_name, args, instance_method = self._process_trigger_handler(
            self.original_app
        )

        return (
            TriggerNode(module, class_name, instance_method, method_name, "RETURN"),
            args,
        )

    async def extract_response_to_context_async(self, response, context):
        """
        Async method to extract response information.
        Unlike the similarly named method in RequestContext, we define
        this method here because defining any async method in RequestContext
        causes SyntaxError for Py2. We can move this once we deprecate Py2.
        """
        context.response = response

        if not self.settings.response_scanning_enabled:
            return

        await response.body_async

    @cached_property
    def name(self):
        return "fastapi"



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import ast
import importlib
import inspect
import sys
import textwrap
from types import FunctionType, ModuleType

from contrast.agent.policy import patch_manager
from contrast.agent.policy.loader import Policy
from contrast.agent.rewriter import PropagationRewriter, populate_rewritten_functions
from contrast.agent.settings import Settings
from contrast.utils.decorators import fail_quietly
from contrast.extern import structlog as logging


logger = logging.getLogger("contrast")


def get_or_load_module(module: ModuleType) -> ModuleType:
    """
    Returns a non-frozen version of the given module

    Frozen modules are modules whose bytecode is built into the interpreter
    itself for performance reasons. All of the modules required at interpreter
    startup are frozen in newer versions of Python (i.e. >= 3.10+).

    Frozen modules do not have access to the source code of module members such
    as functions. This means that calls to inspect.getsource(<frozen-module-name>.<member-name>)
    will fail. Since we need the source in order to perform rewrites, we need a
    non-frozen version of the module. We can achieve this by loading the module
    again and returning the new module object.
    """
    if module.__loader__ is not importlib.machinery.FrozenImporter:
        return module

    temp_name = "__contrast_temp." + module.__name__
    spec = importlib.util.spec_from_file_location(temp_name, module.__file__)

    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)

    assert temp_name not in sys.modules

    return module


@fail_quietly("Failed to apply rewrites to function")
def rewrite_function(module: ModuleType, name: str, source: str):
    """
    Rewrites the function defined in the given module

    The process invovles the following steps:
    1. Retrieve the source for the given function. (This may fail if the function is defined in an extension)
    2. Generate an AST from the original source code
    3. Transform the AST using our PropagationRewriter
    4. Compile the new AST into a code object
    5. Evaluate the new code object with exec
    6. Retrieve the newly defined function from the locals of the exec call
    """
    filename = module.__file__ or "<unknown>"
    mode = "exec"

    old_ast = ast.parse(textwrap.dedent(source), filename=filename, mode=mode)

    rewriter = PropagationRewriter.with_default_settings()
    new_ast = ast.fix_missing_locations(rewriter.visit(old_ast))

    locals_ = {}
    new_code = compile(new_ast, filename, mode=mode)
    exec(new_code, module.__dict__, locals_)

    return locals_[name]


@fail_quietly("Failed rewrite and patch function")
def rewrite_and_patch_function(
    module: ModuleType, name: str, function: FunctionType, source: str
):
    # Some functions may already be patched by other policy, in which case we do not want to rewrite them
    if patch_manager.is_patched(function):
        logger.debug("Skipping rewrite of already patched function: %s", name)
        return

    new_func = rewrite_function(module, name, source)
    if new_func is None:
        logger.debug("No new function for %s. Skipping patch", name)
        return

    patch_manager.patch(module, name, new_func)


@fail_quietly("Failed to rewrite functions for module")
def rewrite_module_functions(module_name: str):
    module = sys.modules.get(module_name, None)
    if module is None:
        logging.debug(
            'Failed to rewrite functions in module "%s": module not loaded', module_name
        )
        return

    logger.debug("Applying rewriter policy to module: %s", module_name)
    temp_module = get_or_load_module(module)

    for name, function in [
        (name, member)
        for name, member in inspect.getmembers(module)
        if inspect.isfunction(member)
    ]:
        # If the unfrozen module doesn't have a function that is found in the
        # original module, it's probably the case that the function was added
        # by us (e.g. a function added by some other policy node).
        if not hasattr(temp_module, name):
            continue

        try:
            source = inspect.getsource(getattr(temp_module, name))
        except Exception:
            logger.debug("No source found for function: %s", name)
            continue

        rewrite_and_patch_function(module, name, function, source)

    populate_rewritten_functions(module.__dict__)


def apply_rewrite_policy():
    if not Settings().is_rewriter_enabled:
        return

    policy = Policy()
    for module in policy.rewrite_modules:
        rewrite_module_functions(module)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
class SecurityException(Exception):

    STATUS_CODE = 403
    REASON_PHRASE = "Forbidden"
    STATUS = f"{STATUS_CODE} {REASON_PHRASE}"

    def __init__(self, rule, message=None):
        message = message if message else f"Rule {rule.name} threw a security exception"
        super(SecurityException, self).__init__(message)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
try:
    from django.http.response import HttpResponseBase

    django_response_types = (HttpResponseBase,)
except ImportError:  # pragma: no cover
    django_response_types = ()

from contrast.extern import structlog as logging

logger = logging.getLogger("contrast")

string_types = (str, bytes, bytearray)

list_iterator = type(iter(list()))
dict_iterator = type(iter(dict()))


def is_iterable(value):
    """
    :param value: any type
    :return: True if iterable type like list, dict but NOT string
             False if string or any other type like database collection types
                   (pymongo Collection implements iter but is not an iterable
    """

    try:
        it = iter(value)
        if not hasattr(it, "__next__"):
            return False
    except TypeError:
        return False

    return not isinstance(value, string_types)


def len_or_zero(value):
    try:
        return len(value)
    except Exception:
        return 0


def safe_iterator(it):
    if isinstance(it, (list_iterator, dict_iterator)):
        logger.debug(
            "WARNING: skipping iteration of non-seekable iterator: %s", type(it)
        )
        return

    if isinstance(it, (str, list, dict, tuple)) or (
        isinstance(it, django_response_types)
        and not safe_getattr(it, "streaming", True)
    ):
        for x in it:
            yield x
        return

    if not (hasattr(it, "tell") and hasattr(it, "seek")):
        logger.debug("WARNING: skipping iteration of non-seekable object: %s", type(it))
        return

    try:
        orig_pos = it.tell()
    except Exception:
        logger.debug(
            "WARNING: skipping iteration of non-tell()-able object: %s", type(it)
        )
        return

    try:
        for x in it:
            yield x
    except Exception:
        logger.debug("safe_iterator failed to iterate over %s", type(it))
    finally:
        it.seek(orig_pos)


def safe_getattr(obj, attr, default=None):
    """
    A getattr implementation that returns a default even if an exception is raised

    Some classes may override __getattribute__ to raise exceptions when certain
    attributes are accessed. When we are dealing with objects from the outside world,
    we want to be careful in our attribute access to avoid such exceptions and provide
    a safe default.
    """
    try:
        return getattr(obj, attr, default)
    except Exception:
        return default



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import os
import sys

from contrast.agent.assess.rules.config import (
    DjangoHttpOnlyRule,
    DjangoSecureFlagRule,
    DjangoSessionAgeRule,
)

from contrast.wsgi.middleware import WSGIMiddleware
from contrast.utils.decorators import fail_quietly
from contrast.agent.middlewares.route_coverage.django_routes import DjangoRoutesMixin
from contrast.extern import structlog as logging
from contrast.utils.decorators import cached_property

logger = logging.getLogger("contrast")


class DjangoWSGIMiddleware(DjangoRoutesMixin, WSGIMiddleware):
    """
    A subclass of the WSGI middleware that provides django route coverage and config
    scanning.

    This is not a Django-style middleware - it must wrap django's WSGI_APPLICATION,
    and does not belong in MIDDLEWARE / MIDDLEWARE_CLASSES.
    """

    def __init__(self, wsgi_app):
        self.app_name = self.get_app_name()

        self.config_rules = [
            DjangoSessionAgeRule(),
            DjangoSecureFlagRule(),
            DjangoHttpOnlyRule(),
        ]

        super().__init__(wsgi_app, self.app_name)

    @fail_quietly(
        "Unable to get Django application name", return_value="Django Application"
    )
    def get_app_name(self):
        from django.conf import settings

        wsgi_application = settings.WSGI_APPLICATION

        return wsgi_application.split(".")[0]

    @fail_quietly("Failed to run config scanning rules")
    def _scan_configs(self):
        """
        Run config scanning rules for assess

        Overridden from base class; gets called from base class
        """
        from django.conf import settings as app_settings

        app_config_module_name = os.environ.get("DJANGO_SETTINGS_MODULE")
        if not app_config_module_name:
            logger.warning("Unable to find Django settings for config scanning")
            return

        app_config_module = sys.modules.get(app_config_module_name)
        if not app_config_module:
            logger.warning("Django settings module not loaded; can't scan config")
            return

        for rule in self.config_rules:
            rule.apply(app_settings, app_config_module)

    @cached_property
    def name(self):
        return "django"



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from starlette.concurrency import iterate_in_threadpool
from contrast.extern.webob.headers import ResponseHeaders
from contrast.agent.middlewares.response_wrappers.base_response_wrapper import (
    BaseAsyncResponseWrapper,
)


class FastApiResponseWrapper(BaseAsyncResponseWrapper):
    def __init__(self, response):
        self._response = response
        self._streaming_cache = None
        self.body = None

    @property
    async def body_async(self):
        # Store body to an attr to be able to retrieve it without calling async code
        self.body = await self._get_streaming_content()

    @property
    def headers(self):
        return ResponseHeaders(self._response.headers)

    @property
    def status_code(self):
        return self._response.status_code

    async def _get_streaming_content(self):
        """
        Safely extract the content of a streaming response body. This method guarantees that the
        response body is restored after extraction. Unfortunately, this is likely a performance hit
        to applications streaming large response bodies.

        :return: body of a streaming response as bytes
        """
        if not self._streaming_cache:
            # body_iterator is a AsyncGenerator
            chunks = [chunk async for chunk in self._response.body_iterator]
            body = b"".join(chunks)
            # The original body_iterator is defined in
            # starlette/middleware/base.py:BaseHTTPMiddleware:call_next:body_stream
            # Which is then passed on to StreamingResponse. The latter calls
            # iterate_in_threadpool, too, so we use that same idea here.
            self._response.body_iterator = iterate_in_threadpool(iter(chunks))
            self._streaming_cache = body
            return self._streaming_cache

        return self._streaming_cache



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import os
from contextlib import contextmanager
import logging as stdlib_logging

import contrast
from contrast.utils.exceptions.security_exception import SecurityException
from contrast.utils.timer import now_ms_float
from contrast.utils.loggers import TRACE_LEVEL
from contrast.extern import structlog as logging
from contrast.agent.validator import ValidationException

logger = logging.getLogger("contrast")

DEBUG_LEVEL = stdlib_logging.DEBUG


def _log_extra_safely(
    log_message: str, error: Exception, original_func, args, kwargs, log_level="debug"
):
    if not contrast.telemetry_disabled():
        TELEMETRY = contrast.TELEMETRY
    else:
        TELEMETRY = None

    try:
        full_msg = log_message + ": " + str(error)
        getattr(logger, log_level)(full_msg, exc_info=error)
        logger.debug("wrapped function args: %s", args)
        logger.debug("wrapped function kwargs: %s", kwargs)

        if TELEMETRY is not None:
            TELEMETRY.report_error(
                error=error, original_func=original_func, message=full_msg
            )
    except ValidationException as val_ex:
        logger.debug(f"Cannot report error to Telemetry: {str(val_ex)}")
    except Exception:
        # For complete safety, we're not going to try to log the logging error
        # because sometimes the logging error is actually an error (such as a recursive
        # error) within the logging module itself!
        pass


def fail_loudly(log_message, log_level="exception", return_value=None):
    """
    Decorator that will run the decorated function/method and, if
    an exception is raised, return a safe value and log the error.

    Note that SecurityException will always be re-raised.

    :param log_message: message to log in case of failure
    :param log_level: level to log in case of failure
    :param return_value: safe value to return in case of failure
    :return: original func return or return_value
    """

    def wrap(original_func):
        def run_safely(*args, **kwargs):
            try:
                return original_func(*args, **kwargs)
            except SecurityException:
                raise
            except Exception as ex:
                _log_extra_safely(
                    log_message, ex, original_func, args, kwargs, log_level
                )
                if os.environ.get("CONTRAST_TESTING"):
                    raise

            return return_value

        return run_safely

    return wrap


def log_time(log_message):
    """
    Decorator that will TRACE log before some method/func runs,
    run the function, log after, and finally log the elapsed time.
    """

    def wrap(original_func):
        def _log_time(*args, **kwargs):
            # Don't do any work if logger is not TRACE.
            if logger.level > TRACE_LEVEL:
                return original_func(*args, **kwargs)

            # Since we can't grab the "time" value from the trace log msg,
            # we compute the time ourselves. This means there will be a tiny,
            # probably insignificant difference between the logging message and the
            # time calculation.
            logger.trace("start %s", log_message)

            start = now_ms_float()
            return_value = original_func(*args, **kwargs)
            end = now_ms_float()

            logger.trace("end %s", log_message)

            elapsed_time = end - start

            logger.trace("elapsed time ms %s", log_message, elapsed_time=elapsed_time)

            return return_value

        return _log_time

    return wrap


@contextmanager
def log_time_cm(log_message):
    """
    Context manager for DEBUG and TRACE level logging.
    If logger is using:
        DEBUG LEVEL: log start, do work, log end at debug level
        TRACE LEVEL: log start, keep time, do work, log end, log elapsed time at
            trace level
        ANY OTHER LEVEL: just do work
    """
    if logger.level > DEBUG_LEVEL:
        yield

    elif logger.level is DEBUG_LEVEL:
        logger.debug("start %s", log_message)
        try:
            yield
        finally:
            logger.debug("end %s", log_message)

    elif logger.level is TRACE_LEVEL:
        logger.trace("start %s", log_message)
        start = now_ms_float()
        try:
            yield
        finally:
            end = now_ms_float()

            logger.trace("end %s", log_message)

            elapsed_time = end - start

            logger.trace("elapsed time ms %s", log_message, elapsed_time=elapsed_time)


def fail_quietly(log_message, return_value=None):
    """
    Similar to fail_loudly (see above)

    This decorator is intended to handle cases where an exception may occur but won't
    disrupt normal operation of the agent. This decorator should be used to protect
    against external exceptions we can't prevent but still want to handle.

    In these cases, we log an error message and the exception traceback, both at DEBUG
    level.
    """

    def wrap(original_func):
        def run_safely(*args, **kwargs):
            try:
                return original_func(*args, **kwargs)
            except SecurityException:
                raise
            except Exception as ex:
                _log_extra_safely(log_message, ex, original_func, args, kwargs)
                if os.environ.get("CONTRAST_TESTING"):
                    raise
            return return_value

        return run_safely

    return wrap


class cached_property(object):
    """
    https://github.com/pydanny/cached-property
    """

    def __init__(self, func):
        self.func = func

    def __get__(self, obj, cls):
        if obj is None:
            return self

        value = obj.__dict__[self.func.__name__] = self.func(obj)
        return value



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.

"""
The sitecustomize.py file is automatically loaded by Python during interpreter initialization.

Our runner script ensures that it is on the PYTHONPATH which is sufficient to make sure it is loaded.

See https://docs.python.org/3/library/site.html for additional details
"""

from contrast.agent.runner import start_runner

start_runner()



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import importlib.abc
import importlib.machinery
import importlib.util
import types
import sys
import ast
import copy


import contrast
from contrast.agent import scope
from contrast.agent.assess.utils import clear_properties, track_and_tag
from contrast.agent.request_context import RequestContext
from contrast.agent.settings import Settings
from contrast.patches.rewriter import contrast__add, populate_rewritten_functions
from contrast.utils.decorators import fail_loudly
from contrast.utils.environ import test_environ
from contrast.extern import structlog as logging
from contrast.extern.wrapt.importer import _ImportHookChainedLoader

logger = logging.getLogger("contrast")


def _load_module(source, module: types.ModuleType, filename):
    """
    Convenience method to compile and execute the given module source

    It seems like we do not need any exception handling here since any
    exception that occurs gets handled further up by the import machinery and
    causes it to fall back on the original loader. This is definitely good for
    us since it means that even if we mess up here somehow, it shouldn't
    prevent the original module from being loaded. It will just be loaded
    without our rewrites.
    """
    code = compile(source, filename, "exec", dont_inherit=True)
    exec(code, module.__dict__)


class RewriterSettings:
    enable_concat: bool = False
    enable_fstring: bool = False
    enable_cformat_modulo: bool = False

    def __init__(
        self,
        enable_concat: bool = False,
        enable_fstring: bool = False,
        enable_cformat_modulo: bool = False,
    ):
        self.enable_concat = enable_concat
        self.enable_fstring = enable_fstring
        self.enable_cformat_modulo = enable_cformat_modulo

    @classmethod
    def default(cls, override_check: bool = False):
        enable_concat = override_check or not concat_works()
        enable_fstring = override_check or not fstring_works()
        enable_cformat_modulo = override_check or not cformat_modulo_works()

        return cls(
            enable_concat=enable_concat,
            enable_fstring=enable_fstring,
            enable_cformat_modulo=enable_cformat_modulo,
        )


class ContrastMetaPathFinder(importlib.abc.MetaPathFinder):
    settings: RewriterSettings

    def __init__(self, settings: RewriterSettings):
        self.settings = settings

    @fail_loudly("Unexpected error in find_spec - will not rewrite this module")
    def find_spec(self, fullname, path, target=None):
        """
        The finder is in charge of finding a module's "spec". The spec includes import
        machinery metadata about the module - including its name, source file path, and
        the loader, among others.

        Here, we first use importlib's default machinery to get the spec for the module
        about to be imported. The problem with this spec is that it also uses the
        default loader, which isn't what we want. To get around this, we reuse some
        metadata and generate a new spec that points at our loader.

        It's possible that this is needlessly complicated. It's a first-pass
        implementation, so we can (and should) refactor this as we learn more.
        """
        default_spec = importlib.machinery.PathFinder.find_spec(fullname, path)

        if not default_spec:
            logger.debug(
                "WARNING: no spec found for module - fullname=<%s>, path=<%s>",
                fullname,
                path,
            )
            return None

        default_spec_origin = getattr(default_spec, "origin", "") or ""
        if not default_spec_origin.endswith(".py"):
            logger.debug(
                "Will not rewrite non *.py file - fullname=<%s>, path=<%s>",
                fullname,
                default_spec_origin,
            )
            return None

        spec = importlib.util.spec_from_file_location(
            fullname,
            default_spec.origin,
            loader=ContrastRewriteLoader(self.settings),
            submodule_search_locations=default_spec.submodule_search_locations,
        )
        if spec is None:
            return spec

        loader = getattr(spec, "loader", None)

        if loader and not isinstance(loader, _ImportHookChainedLoader):
            spec.loader = _ImportHookChainedLoader(loader)

        return spec


class ContrastRewriteLoader(importlib.abc.Loader):
    settings: RewriterSettings

    def __init__(self, settings: RewriterSettings):
        self.settings = settings

    def create_module(self, _):
        """returning None uses the default behavior, which is fine"""
        return None

    def exec_module(self, module: types.ModuleType) -> None:
        """
        This method is responsible for actually doing the module `exec`-ing. We take
        control of this system and do the following:
        - read the original source file. We require pyc caching to be disabled for this
        - parse the source file into an AST
        - rewrite the AST
        - compile the AST into a code object
        - exec the code object

        Note that we add our custom add function to the module's globals. This prevents
        the need for import rewriting entirely
        """
        with scope.contrast_scope():
            original_source_code = None
            filename = module.__spec__.origin

            # May be None in some cases such as for namespace packages
            if filename is None:
                return

            try:
                with open(filename) as f:
                    original_source_code = f.read()

                tree = ast.parse(original_source_code)
            except Exception as ex:
                logger.debug("WARNING: failed to rewrite %s", filename, exc_info=ex)

                _load_module(original_source_code, module, filename)

                return

            if contrast__add.__name__ not in module.__dict__:
                try:
                    populate_rewritten_functions(module.__dict__)
                    PropagationRewriter(self.settings).visit(tree)
                    ast.fix_missing_locations(tree)
                except Exception as ex:
                    logger.debug("WARNING: failed to rewrite %s", filename, exc_info=ex)
            else:
                logger.debug(
                    "WARNING: contrast__add is already defined in %s; will not rewrite",
                    filename,
                )

            _load_module(tree, module, filename)


class PropagationRewriter(ast.NodeTransformer):
    settings: RewriterSettings

    @classmethod
    def with_default_settings(cls, override_check: bool = False):
        return cls(settings=RewriterSettings.default(override_check=override_check))

    def __init__(self, settings: RewriterSettings):
        self.settings = settings

    def _with_context(self, node, context):
        node = copy.copy(node)
        node.ctx = context
        return node

    def visit_BinOp(self, binop: ast.BinOp):
        """
        If we see an "Add" or a "Mod" binary operation, replace it with a call to our custom add/modulo
        function, which includes all necessary instrumentation.
        """
        binop.left = self.visit(binop.left)
        binop.right = self.visit(binop.right)

        if self.settings.enable_cformat_modulo and isinstance(binop.op, ast.Mod):
            binop_replacement = ast.Call(
                func=ast.Name(id="contrast__cformat__modulo", ctx=ast.Load()),
                args=[binop.left, binop.right],
                keywords=[],
            )

            ast.copy_location(binop_replacement, binop)
            return binop_replacement

        if not self.settings.enable_concat:
            return binop

        if not isinstance(binop.op, ast.Add):
            return binop

        binop_replacement = ast.Call(
            func=ast.Name(id="contrast__add", ctx=ast.Load()),
            args=[binop.left, binop.right],
            keywords=[],
        )
        ast.copy_location(binop_replacement, binop)
        return binop_replacement

    def visit_AugAssign(self, node: ast.AugAssign):
        """
        If we see an "Append", `+=` operation, rewrite it as a `+`.
        """
        node.value = self.visit(node.value)

        if not self.settings.enable_concat:
            return node

        if not isinstance(node.op, ast.Add):
            return node

        target = left = None
        if isinstance(node.target, ast.Name):
            name = ast.Name(id=node.target.id)
            target = self._with_context(name, ast.Store())
            left = self._with_context(name, ast.Load())
        else:
            target = node.target
            left = self._with_context(target, ast.Load())

        call_contrast_append_node = ast.Assign(
            targets=[target],
            value=ast.Call(
                func=ast.Name(id="contrast__append", ctx=ast.Load()),
                args=[self.visit(left), node.value],
                keywords=[],
            ),
        )

        ast.copy_location(call_contrast_append_node, node)
        # Recurse here in order to properly rewrite the `+` operation
        return call_contrast_append_node

    def visit_JoinedStr(self, node: ast.JoinedStr):
        if not self.settings.enable_fstring:
            return node

        return ast.Call(
            func=ast.Name(id="contrast__fstring", ctx=ast.Load()),
            args=node.values,
            keywords=[],
        )


@fail_loudly("Unexpected error concat_works", log_level="debug", return_value=False)
def concat_works():
    """
    Tests to see if concat via hooks patches works in this environment
    by running a str concat test as if we were in a request.

    Return: True if propagation happened, False otherwise.

    TODO: PYT-2678 Remove when concat hooks are finally removed
    """
    logger.debug("Testing if concat works.")

    with scope.pop_contrast_scope():
        one = "one"
        two = "two"
        track_and_tag(one)

        context = RequestContext(test_environ)
        with contrast.CS__CONTEXT_TRACKER.lifespan(context):
            res = one + two

    properties = contrast.STRING_TRACKER.get(res)
    result = properties is not None

    clear_properties()
    return result


@fail_loudly(
    "Unexpected error cformat_modulo_works", log_level="debug", return_value=False
)
def cformat_modulo_works():
    """
    Tests to see if modulo string/bytes formatting via hooks patches works in this environment
    by running a test as if we were in a request.

    Return: True if propagation happened, False otherwise.

    TODO: PYT-2710 Remove when cformat hooks are finally removed
    """
    logger.debug("Testing if modulo string/byte formatting works.")

    with scope.pop_contrast_scope():
        fmt = "%s %s"
        args = ("foo", "bar")
        track_and_tag(args[0])

        context = RequestContext(test_environ)
        with contrast.CS__CONTEXT_TRACKER.lifespan(context):
            res = fmt % args  # pylint: disable=string-format-interpolation

    properties = contrast.STRING_TRACKER.get(res)
    result = properties is not None

    clear_properties()
    return result


@fail_loudly("Unexpected error fstring_works", log_level="debug", return_value=False)
def fstring_works():
    """
    TODO: PYT-2679 Remove when fstring hooks are finally removed
    """
    logger.debug("Testing if fstring works.")

    with scope.pop_contrast_scope():
        string = "whatever"
        track_and_tag(string)

        context = RequestContext(test_environ)
        with contrast.CS__CONTEXT_TRACKER.lifespan(context):
            result = f"hello {string}!"

    properties = contrast.STRING_TRACKER.get(result)
    result = properties is not None

    clear_properties()
    return result


def register(override_settings=False, override_check=False):
    """
    Register our rewriter with the import system. After this call, any newly imported
    modules (from source code) will use our custom rewriter.

    Note that because this function is defined in the same module that defines our add
    replacement function, we never have to worry about rewriting the addition in the
    replacement function itself. If that were to occur, we would get an infinite
    recursion.

    Rewriter should only run in >=py3.10 and only in environments in which our default
    patching mechanism does not work.

    :param override_settings: Force the rewriter to be registered regardless of settings (for testing purposes)
    """
    settings = Settings()

    if not (settings.is_rewriter_enabled or override_settings):
        logger.debug("Rewriter disabled exiting setup")
        return

    rewriter_settings = RewriterSettings.default(override_check=override_check)
    sys.meta_path.insert(0, ContrastMetaPathFinder(rewriter_settings))

    settings.rewriter_enabled = True

    logger.debug("enabled AST rewriter")


def deregister():
    """
    Remove our rewriter from the import system. Modules that were loaded by our rewriter
    will remain rewritten.

    Return True if we find and deregister our machinery, False otherwise.
    """
    for i, finder in enumerate(sys.meta_path.copy()):
        if isinstance(finder, ContrastMetaPathFinder):
            sys.meta_path.pop(i)
            # Do we need to set settings.rewriter_enabled = False here?
            return True
    return False



# SPDX-License-Identifier: MIT OR Apache-2.0
# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the MIT License.  See the LICENSE file in the root of this
# repository for complete details.

"""
Processors useful regardless of the logging framework.
"""

from __future__ import annotations

import datetime
import enum
import inspect
import json
import logging
import operator
import os
import sys
import threading
import time

from typing import (
    Any,
    Callable,
    ClassVar,
    Collection,
    NamedTuple,
    Sequence,
    TextIO,
)

from ._frames import (
    _find_first_app_frame_and_name,
    _format_exception,
    _format_stack,
)
from ._log_levels import _NAME_TO_LEVEL, add_log_level
from ._utils import get_processname
from .tracebacks import ExceptionDictTransformer
from .typing import EventDict, ExceptionTransformer, ExcInfo, WrappedLogger


__all__ = [
    "_NAME_TO_LEVEL",  # some people rely on it being here
    "add_log_level",
    "CallsiteParameter",
    "CallsiteParameterAdder",
    "dict_tracebacks",
    "EventRenamer",
    "ExceptionPrettyPrinter",
    "format_exc_info",
    "JSONRenderer",
    "KeyValueRenderer",
    "StackInfoRenderer",
    "TimeStamper",
    "UnicodeDecoder",
    "UnicodeEncoder",
]


class KeyValueRenderer:
    """
    Render ``event_dict`` as a list of ``Key=repr(Value)`` pairs.

    :param sort_keys: Whether to sort keys when formatting.
    :param key_order: List of keys that should be rendered in this exact
        order.  Missing keys will be rendered as ``None``, extra keys depending
        on *sort_keys* and the dict class.
    :param drop_missing: When ``True``, extra keys in *key_order* will be
        dropped rather than rendered as ``None``.
    :param repr_native_str: When ``True``, :func:`repr()` is also applied
        to native strings.
        Setting this to ``False`` is useful if you want to have human-readable
        non-ASCII output on Python 2.

    .. versionadded:: 0.2.0 *key_order*
    .. versionadded:: 16.1.0 *drop_missing*
    .. versionadded:: 17.1.0 *repr_native_str*
    """

    def __init__(
        self,
        sort_keys: bool = False,
        key_order: Sequence[str] | None = None,
        drop_missing: bool = False,
        repr_native_str: bool = True,
    ):
        self._ordered_items = _items_sorter(sort_keys, key_order, drop_missing)

        if repr_native_str is True:
            self._repr = repr
        else:

            def _repr(inst: Any) -> str:
                if isinstance(inst, str):
                    return inst
                else:
                    return repr(inst)

            self._repr = _repr

    def __call__(
        self, _: WrappedLogger, __: str, event_dict: EventDict
    ) -> str:
        return " ".join(
            k + "=" + self._repr(v) for k, v in self._ordered_items(event_dict)
        )


class LogfmtRenderer:
    """
    Render ``event_dict`` using the logfmt_ format.

    .. _logfmt: https://brandur.org/logfmt

    :param sort_keys: Whether to sort keys when formatting.
    :param key_order: List of keys that should be rendered in this exact
        order. Missing keys are rendered with empty values, extra keys
        depending on *sort_keys* and the dict class.
    :param drop_missing: When ``True``, extra keys in *key_order* will be
        dropped rather than rendered with empty values.
    :param bool_as_flag: When ``True``, render ``{"flag": True}`` as
        ``flag``, instead of ``flag=true``. ``{"flag": False}`` is
        always rendered as ``flag=false``.

    :raises ValueError: If a key contains non printable or space characters.

    .. versionadded:: 21.5.0
    """

    def __init__(
        self,
        sort_keys: bool = False,
        key_order: Sequence[str] | None = None,
        drop_missing: bool = False,
        bool_as_flag: bool = True,
    ):
        self._ordered_items = _items_sorter(sort_keys, key_order, drop_missing)
        self.bool_as_flag = bool_as_flag

    def __call__(
        self, _: WrappedLogger, __: str, event_dict: EventDict
    ) -> str:

        elements: list[str] = []
        for key, value in self._ordered_items(event_dict):
            if any(c <= " " for c in key):
                raise ValueError(f'Invalid key: "{key}"')

            if value is None:
                elements.append(f"{key}=")
                continue

            if isinstance(value, bool):
                if self.bool_as_flag and value:
                    elements.append(f"{key}")
                    continue
                value = "true" if value else "false"

            value = f"{value}".replace('"', '\\"')

            if " " in value or "=" in value:
                value = f'"{value}"'

            elements.append(f"{key}={value}")

        return " ".join(elements)


def _items_sorter(
    sort_keys: bool,
    key_order: Sequence[str] | None,
    drop_missing: bool,
) -> Callable[[EventDict], list[tuple[str, Any]]]:
    """
    Return a function to sort items from an ``event_dict``.

    See `KeyValueRenderer` for an explanation of the parameters.
    """
    # Use an optimized version for each case.
    if key_order and sort_keys:

        def ordered_items(event_dict: EventDict) -> list[tuple[str, Any]]:
            items = []
            for key in key_order:  # type: ignore[union-attr]
                value = event_dict.pop(key, None)
                if value is not None or not drop_missing:
                    items.append((key, value))

            items += sorted(event_dict.items())

            return items

    elif key_order:

        def ordered_items(event_dict: EventDict) -> list[tuple[str, Any]]:
            items = []
            for key in key_order:  # type: ignore[union-attr]
                value = event_dict.pop(key, None)
                if value is not None or not drop_missing:
                    items.append((key, value))

            items += event_dict.items()

            return items

    elif sort_keys:

        def ordered_items(event_dict: EventDict) -> list[tuple[str, Any]]:
            return sorted(event_dict.items())

    else:
        ordered_items = operator.methodcaller(  # type: ignore[assignment]
            "items"
        )

    return ordered_items


class UnicodeEncoder:
    """
    Encode unicode values in ``event_dict``.

    :param encoding: Encoding to encode to (default: ``"utf-8"``).
    :param errors: How to cope with encoding errors (default
        ``"backslashreplace"``).

    Useful if you're running Python 2 as otherwise ``u"abc"`` will be rendered
    as ``'u"abc"'``.

    Just put it in the processor chain before the renderer.
    """

    _encoding: str
    _errors: str

    def __init__(
        self, encoding: str = "utf-8", errors: str = "backslashreplace"
    ) -> None:
        self._encoding = encoding
        self._errors = errors

    def __call__(
        self, logger: WrappedLogger, name: str, event_dict: EventDict
    ) -> EventDict:
        for key, value in event_dict.items():
            if isinstance(value, str):
                event_dict[key] = value.encode(self._encoding, self._errors)

        return event_dict


class UnicodeDecoder:
    """
    Decode byte string values in ``event_dict``.

    :param encoding: Encoding to decode from (default: ``"utf-8"``).
    :param errors: How to cope with encoding errors (default:
        ``"replace"``).

    Useful if you're running Python 3 as otherwise ``b"abc"`` will be rendered
    as ``'b"abc"'``.

    Just put it in the processor chain before the renderer.

    .. versionadded:: 15.4.0
    """

    _encoding: str
    _errors: str

    def __init__(
        self, encoding: str = "utf-8", errors: str = "replace"
    ) -> None:
        self._encoding = encoding
        self._errors = errors

    def __call__(
        self, logger: WrappedLogger, name: str, event_dict: EventDict
    ) -> EventDict:
        for key, value in event_dict.items():
            if isinstance(value, bytes):
                event_dict[key] = value.decode(self._encoding, self._errors)

        return event_dict


class JSONRenderer:
    """
    Render the ``event_dict`` using ``serializer(event_dict, **dumps_kw)``.

    :param dumps_kw: Are passed unmodified to *serializer*.  If *default*
        is passed, it will disable support for ``__structlog__``-based
        serialization.
    :param serializer: A :func:`json.dumps`-compatible callable that
        will be used to format the string.  This can be used to use alternative
        JSON encoders like `orjson <https://pypi.org/project/orjson/>`__ or
        `RapidJSON <https://pypi.org/project/python-rapidjson/>`_  (default:
        :func:`json.dumps`).

    .. versionadded:: 0.2.0
        Support for ``__structlog__`` serialization method.

    .. versionadded:: 15.4.0
        *serializer* parameter.

    .. versionadded:: 18.2.0
       Serializer's *default* parameter can be overwritten now.

    """

    def __init__(
        self,
        serializer: Callable[..., str | bytes] = json.dumps,
        **dumps_kw: Any,
    ) -> None:
        dumps_kw.setdefault("default", _json_fallback_handler)
        self._dumps_kw = dumps_kw
        self._dumps = serializer

    def __call__(
        self, logger: WrappedLogger, name: str, event_dict: EventDict
    ) -> str | bytes:
        """
        The return type of this depends on the return type of self._dumps.
        """
        return self._dumps(event_dict, **self._dumps_kw)


def _json_fallback_handler(obj: Any) -> Any:
    """
    Serialize custom datatypes and pass the rest to __structlog__ & repr().
    """
    # circular imports :(
    from contrast.extern.structlog.threadlocal import _ThreadLocalDictWrapper

    if isinstance(obj, _ThreadLocalDictWrapper):
        return obj._dict
    else:
        try:
            return obj.__structlog__()
        except AttributeError:
            return repr(obj)


class ExceptionRenderer:
    """
    Replace an ``exc_info`` field with an ``exception`` field which is rendered
    by *exception_formatter*.

    The contents of the ``exception`` field depends on the return value of the
    :class:`.ExceptionTransformer` that is used:

    - The default produces a formatted string via Python's built-in traceback
      formatting.
    - The :class:`~structlog.tracebacks.ExceptionDictTransformer` a list of
      stack dicts that can be serialized to JSON.

    If *event_dict* contains the key ``exc_info``, there are three possible
    behaviors:

    1. If the value is a tuple, render it into the key ``exception``.
    2. If the value is an Exception render it into the key ``exception``.
    3. If the value true but no tuple, obtain exc_info ourselves and render
       that.

    If there is no ``exc_info`` key, the *event_dict* is not touched.
    This behavior is analogue to the one of the stdlib's logging.

    :param exception_formatter: A callable that is used to format the exception
        from the ``exc_info`` field.

    .. versionadded:: 22.1
    """

    def __init__(
        self,
        exception_formatter: ExceptionTransformer = _format_exception,
    ) -> None:
        self.format_exception = exception_formatter

    def __call__(
        self, logger: WrappedLogger, name: str, event_dict: EventDict
    ) -> EventDict:
        exc_info = event_dict.pop("exc_info", None)
        if exc_info:
            event_dict["exception"] = self.format_exception(
                _figure_out_exc_info(exc_info)
            )

        return event_dict


format_exc_info = ExceptionRenderer()
"""
Replace an ``exc_info`` field with an ``exception`` string field using
Python's built-in traceback formatting.

If *event_dict* contains the key ``exc_info``, there are three possible
behaviors:

1. If the value is a tuple, render it into the key ``exception``.
2. If the value is an Exception render it into the key ``exception``.
3. If the value is true but no tuple, obtain exc_info ourselves and render
   that.

If there is no ``exc_info`` key, the *event_dict* is not touched.
This behavior is analogue to the one of the stdlib's logging.
"""

dict_tracebacks = ExceptionRenderer(ExceptionDictTransformer())
"""
Replace an ``exc_info`` field with an ``exception`` field containing structured
tracebacks suitable for, e.g., JSON output.

It is a shortcut for :class:`ExceptionRenderer` with a
:class:`~structlog.tracebacks.ExceptionDictTransformer`.

The treatment of the ``exc_info`` key is identical to `format_exc_info`.

.. versionadded:: 22.1
"""


class TimeStamper:
    """
    Add a timestamp to ``event_dict``.

    :param fmt: strftime format string, or ``"iso"`` for `ISO 8601
        <https://en.wikipedia.org/wiki/ISO_8601>`_, or `None` for a `UNIX
        timestamp <https://en.wikipedia.org/wiki/Unix_time>`_.
    :param utc: Whether timestamp should be in UTC or local time.
    :param key: Target key in *event_dict* for added timestamps.

    .. versionchanged:: 19.2 Can be pickled now.
    """

    __slots__ = ("_stamper", "fmt", "utc", "key")

    def __init__(
        self,
        fmt: str | None = None,
        utc: bool = True,
        key: str = "timestamp",
    ) -> None:
        self.fmt, self.utc, self.key = fmt, utc, key

        self._stamper = _make_stamper(fmt, utc, key)

    def __call__(
        self, logger: WrappedLogger, name: str, event_dict: EventDict
    ) -> EventDict:
        return self._stamper(event_dict)

    def __getstate__(self) -> dict[str, Any]:
        return {"fmt": self.fmt, "utc": self.utc, "key": self.key}

    def __setstate__(self, state: dict[str, Any]) -> None:
        self.fmt = state["fmt"]
        self.utc = state["utc"]
        self.key = state["key"]

        self._stamper = _make_stamper(**state)


def _make_stamper(
    fmt: str | None, utc: bool, key: str
) -> Callable[[EventDict], EventDict]:
    """
    Create a stamper function.
    """
    if fmt is None and not utc:
        raise ValueError("UNIX timestamps are always UTC.")

    now: Callable[[], datetime.datetime]

    if utc:

        def now() -> datetime.datetime:
            return datetime.datetime.utcnow()

    else:

        def now() -> datetime.datetime:
            return datetime.datetime.now()

    if fmt is None:

        def stamper_unix(event_dict: EventDict) -> EventDict:
            event_dict[key] = time.time()

            return event_dict

        return stamper_unix
    elif fmt.upper() == "ISO":

        def stamper_iso_local(event_dict: EventDict) -> EventDict:
            event_dict[key] = now().isoformat()
            return event_dict

        def stamper_iso_utc(event_dict: EventDict) -> EventDict:
            event_dict[key] = now().isoformat() + "Z"
            return event_dict

        if utc:
            return stamper_iso_utc

        return stamper_iso_local

    def stamper_fmt(event_dict: EventDict) -> EventDict:
        event_dict[key] = now().strftime(fmt)  # type: ignore[arg-type]

        return event_dict

    return stamper_fmt


def _figure_out_exc_info(v: Any) -> ExcInfo:
    """
    Depending on the Python version will try to do the smartest thing possible
    to transform *v* into an ``exc_info`` tuple.
    """
    if isinstance(v, BaseException):
        return (v.__class__, v, v.__traceback__)
    elif isinstance(v, tuple):
        return v  # type: ignore[return-value]
    elif v:
        return sys.exc_info()  # type: ignore[return-value]

    return v


class ExceptionPrettyPrinter:
    """
    Pretty print exceptions and remove them from the ``event_dict``.

    :param file: Target file for output (default: ``sys.stdout``).

    This processor is mostly for development and testing so you can read
    exceptions properly formatted.

    It behaves like `format_exc_info` except it removes the exception
    data from the event dictionary after printing it.

    It's tolerant to having `format_exc_info` in front of itself in the
    processor chain but doesn't require it.  In other words, it handles both
    ``exception`` as well as ``exc_info`` keys.

    .. versionadded:: 0.4.0

    .. versionchanged:: 16.0.0
       Added support for passing exceptions as ``exc_info`` on Python 3.
    """

    def __init__(
        self,
        file: TextIO | None = None,
        exception_formatter: ExceptionTransformer = _format_exception,
    ) -> None:
        if file is not None:
            self._file = file
        else:
            self._file = sys.stdout

    def __call__(
        self, logger: WrappedLogger, name: str, event_dict: EventDict
    ) -> EventDict:
        exc = event_dict.pop("exception", None)
        if exc is None:
            exc_info = _figure_out_exc_info(event_dict.pop("exc_info", None))
            if exc_info:
                exc = _format_exception(exc_info)

        if exc:
            print(exc, file=self._file)

        return event_dict


class StackInfoRenderer:
    """
    Add stack information with key ``stack`` if ``stack_info`` is `True`.

    Useful when you want to attach a stack dump to a log entry without
    involving an exception and works analogously to the *stack_info* argument
    of the Python standard library logging.

    :param additional_ignores: By default, stack frames coming from
        *structlog* are ignored. With this argument you can add additional
        names that are ignored, before the stack starts being rendered. They
        are matched using ``startswith()``, so they don't have to match
        exactly. The names are used to find the first relevant name, therefore
        once a frame is found that doesn't start with *structlog* or one of
        *additional_ignores*, **no filtering** is applied to subsequent frames.

    .. versionadded:: 0.4.0
    .. versionadded:: 22.1.0  *additional_ignores*
    """

    __slots__ = ["_additional_ignores"]

    def __init__(self, additional_ignores: list[str] | None = None) -> None:
        self._additional_ignores = additional_ignores

    def __call__(
        self, logger: WrappedLogger, name: str, event_dict: EventDict
    ) -> EventDict:
        if event_dict.pop("stack_info", None):
            event_dict["stack"] = _format_stack(
                _find_first_app_frame_and_name(self._additional_ignores)[0]
            )

        return event_dict


class CallsiteParameter(enum.Enum):
    """
    Callsite parameters that can be added to an event dictionary with the
    `structlog.processors.CallsiteParameterAdder` processor class.

    The string values of the members of this enum will be used as the keys for
    the callsite parameters in the event dictionary.

    .. versionadded:: 21.5.0
    """

    #: The full path to the python source file of the callsite.
    PATHNAME = "pathname"
    #: The basename part of the full path to the python source file of the
    #: callsite.
    FILENAME = "filename"
    #: The python module the callsite was in. This mimics the module attribute
    #: of `logging.LogRecord` objects and will be the basename, without
    #: extension, of the full path to the python source file of the callsite.
    MODULE = "module"
    #: The name of the function that the callsite was in.
    FUNC_NAME = "func_name"
    #: The line number of the callsite.
    LINENO = "lineno"
    #: The ID of the thread the callsite was executed in.
    THREAD = "thread"
    #: The name of the thread the callsite was executed in.
    THREAD_NAME = "thread_name"
    #: The ID of the process the callsite was executed in.
    PROCESS = "process"
    #: The name of the process the callsite was executed in.
    PROCESS_NAME = "process_name"


class CallsiteParameterAdder:
    """
    Adds parameters of the callsite that an event dictionary originated from to
    the event dictionary. This processor can be used to enrich events
    dictionaries with information such as the function name, line number and
    filename that an event dictionary originated from.

    .. warning::
        This processor cannot detect the correct callsite for invocation of
        async functions.

    If the event dictionary has an embedded `logging.LogRecord` object and did
    not originate from *structlog* then the callsite information will be
    determined from the `logging.LogRecord` object. For event dictionaries
    without an embedded `logging.LogRecord` object the callsite will be
    determined from the stack trace, ignoring all intra-structlog calls, calls
    from the `logging` module, and stack frames from modules with names that
    start with values in ``additional_ignores``, if it is specified.

    The keys used for callsite parameters in the event dictionary are the
    string values of `CallsiteParameter` enum members.

    :param parameters:
        A collection of `CallsiteParameter` values that should be added to the
        event dictionary.

    :param additional_ignores:
        Additional names with which a stack frame's module name must not
        start for it to be considered when determening the callsite.

    .. note::
        When used with `structlog.stdlib.ProcessorFormatter` the most efficient
        configuration is to either use this processor in ``foreign_pre_chain``
        of `structlog.stdlib.ProcessorFormatter` and in ``processors`` of
        `structlog.configure`, or to use it in ``processors`` of
        `structlog.stdlib.ProcessorFormatter` without using it in
        ``processors`` of `structlog.configure` and ``foreign_pre_chain`` of
        `structlog.stdlib.ProcessorFormatter`.

    .. versionadded:: 21.5.0
    """

    _handlers: ClassVar[
        dict[CallsiteParameter, Callable[[str, inspect.Traceback], Any]]
    ] = {
        CallsiteParameter.PATHNAME: (
            lambda module, frame_info: frame_info.filename
        ),
        CallsiteParameter.FILENAME: (
            lambda module, frame_info: os.path.basename(frame_info.filename)
        ),
        CallsiteParameter.MODULE: (
            lambda module, frame_info: os.path.splitext(
                os.path.basename(frame_info.filename)
            )[0]
        ),
        CallsiteParameter.FUNC_NAME: (
            lambda module, frame_info: frame_info.function
        ),
        CallsiteParameter.LINENO: (
            lambda module, frame_info: frame_info.lineno
        ),
        CallsiteParameter.THREAD: (
            lambda module, frame_info: threading.get_ident()
        ),
        CallsiteParameter.THREAD_NAME: (
            lambda module, frame_info: threading.current_thread().name
        ),
        CallsiteParameter.PROCESS: (lambda module, frame_info: os.getpid()),
        CallsiteParameter.PROCESS_NAME: (
            lambda module, frame_info: get_processname()
        ),
    }
    _record_attribute_map: ClassVar[dict[CallsiteParameter, str]] = {
        CallsiteParameter.PATHNAME: "pathname",
        CallsiteParameter.FILENAME: "filename",
        CallsiteParameter.MODULE: "module",
        CallsiteParameter.FUNC_NAME: "funcName",
        CallsiteParameter.LINENO: "lineno",
        CallsiteParameter.THREAD: "thread",
        CallsiteParameter.THREAD_NAME: "threadName",
        CallsiteParameter.PROCESS: "process",
        CallsiteParameter.PROCESS_NAME: "processName",
    }

    _all_parameters: ClassVar[set[CallsiteParameter]] = set(CallsiteParameter)

    class _RecordMapping(NamedTuple):
        event_dict_key: str
        record_attribute: str

    __slots__ = [
        "_active_handlers",
        "_additional_ignores",
        "_record_mappings",
    ]

    def __init__(
        self,
        parameters: Collection[CallsiteParameter] = _all_parameters,
        additional_ignores: list[str] | None = None,
    ) -> None:
        if additional_ignores is None:
            additional_ignores = []
        # Ignore stack frames from the logging module. They will occur if this
        # processor is used in ProcessorFormatter, and additionally the logging
        # module should not be logging using structlog.
        self._additional_ignores = ["logging", *additional_ignores]
        self._active_handlers: list[
            tuple[CallsiteParameter, Callable[[str, inspect.Traceback], Any]]
        ] = []
        self._record_mappings: list[CallsiteParameterAdder._RecordMapping] = []
        for parameter in parameters:
            self._active_handlers.append(
                (parameter, self._handlers[parameter])
            )
            self._record_mappings.append(
                self._RecordMapping(
                    parameter.value,
                    self._record_attribute_map[parameter],
                )
            )

    def __call__(
        self, logger: logging.Logger, name: str, event_dict: EventDict
    ) -> EventDict:
        record: logging.LogRecord | None = event_dict.get("_record")
        from_structlog: bool | None = event_dict.get("_from_structlog")
        # If the event dictionary has a record, but it comes from structlog,
        # then the callsite parameters of the record will not be correct.
        if record is not None and not from_structlog:
            for mapping in self._record_mappings:
                event_dict[mapping.event_dict_key] = record.__dict__[
                    mapping.record_attribute
                ]
        else:
            frame, module = _find_first_app_frame_and_name(
                additional_ignores=self._additional_ignores
            )
            frame_info = inspect.getframeinfo(frame)
            for parameter, handler in self._active_handlers:
                event_dict[parameter.value] = handler(module, frame_info)
        return event_dict


class EventRenamer:
    r"""
    Rename the ``event`` key in event dicts.

    This is useful if you want to use consistent log message keys across
    platforms and/or use the ``event`` key for something custom.

    .. warning::

       It's recommended to put this processor right before the renderer, since
       some processors may rely on the presence and meaning of the ``event``
       key.

    :param to: Rename ``event_dict["event"]`` to ``event_dict[to]``
    :param replace_by: Rename ``event_dict[replace_by]`` to
        ``event_dict["event"]``. *replace_by* missing from ``event_dict`` is
        handled gracefully.

    .. versionadded:: 22.1

    See also the :ref:`rename-event` recipe.
    """

    def __init__(self, to: str, replace_by: str | None = None):
        self.to = to
        self.replace_by = replace_by

    def __call__(
        self, logger: logging.Logger, name: str, event_dict: EventDict
    ) -> EventDict:
        event = event_dict.pop("event")
        event_dict[self.to] = event

        if self.replace_by is not None:
            replace_by = event_dict.pop(self.replace_by, None)
            if replace_by is not None:
                event_dict["event"] = replace_by

        return event_dict



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.asgi.middleware import ASGIMiddleware
from contrast.agent.middlewares.app_finder import get_original_app_or_fail
from contrast.utils.decorators import cached_property
from contrast.agent.middlewares.route_coverage.falcon_routes import FalconRoutesMixin

# pylint: disable=dangerous-framework-import
import falcon


class FalconAsyncMiddleware(FalconRoutesMixin, ASGIMiddleware):
    def __init__(self, app, orig_falcon_api_instance=None):
        falcon_app = falcon.asgi.app.App

        _app = (
            orig_falcon_api_instance
            if orig_falcon_api_instance is not None
            and isinstance(orig_falcon_api_instance, falcon_app)
            else get_original_app_or_fail(app, falcon_app)
        )
        self.falcon_app = _app
        # used for route coverage only for falcon middleware
        self.endpoint_cls = None

        super().__init__(app, app_name="Falcon Application")

    @cached_property
    def name(self):
        return "falcon"



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import sys

import contrast
from contrast.extern.wrapt import register_post_import_hook
from contrast.agent.assess.policy.analysis import analyze
from contrast.agent.policy import patch_manager
from contrast.agent.policy.applicator import apply_assess_patch
from contrast.agent.policy.loader import Policy
from contrast.patches import urllib_patch
from contrast.utils.decorators import fail_quietly
from contrast.utils.patch_utils import build_and_apply_patch

PYRAMID_SESSION_MODULE = "pyramid.session"


@fail_quietly("Failed to apply policy to new session class")
def _apply_policy(session_cls):
    policy = Policy()

    for patch_policy in policy.policy_by_module[PYRAMID_SESSION_MODULE]:
        if patch_policy.class_name == "CookieSession":
            apply_assess_patch(session_cls, patch_policy)


@fail_quietly("Failed to apply assess policy for BaseCookieSessionFactory")
def _apply_assess(result, args, kwargs):
    context = contrast.CS__CONTEXT_TRACKER.current()

    policy = Policy()
    patch_policy = policy.policy_by_name["pyramid.session.BaseCookieSessionFactory"]

    analyze(context, patch_policy, result, args, kwargs)


def build_base_cookie_session_factory_patch(orig_func, patch_policy):
    def base_cookie_session_factory(*args, **kwargs):
        """
        base_cookie_session_factory is a function that returns a new CookieSession class

        Since we can't instrument the new class directly using normal policy machinery,
        we instead apply our policy on-demand to the newly created class.
        """
        session_cls = None

        try:
            session_cls = orig_func(*args, **kwargs)
            _apply_policy(session_cls)
        finally:
            _apply_assess(session_cls, args, kwargs)

        return session_cls

    return base_cookie_session_factory


def patch_session_pyramid(pyramid_session_module):
    build_and_apply_patch(
        pyramid_session_module,
        "BaseCookieSessionFactory",
        build_base_cookie_session_factory_patch,
    )


def patch_encode_pyramid(pyramid_encode_module):
    # We go ahead and apply all urllib patches here (even though policy
    # patches will happen later on) because we MUST have some urllib policy patches
    # already applied for these non-policy patches to work.
    # This would not be necessary if in _enable_patches policy_patches were applied
    # first.
    urllib_patch.register_patches()

    # We can reuse the urllib.parse.urlencode patch since it's exactly the same
    # as the pyramid.encode.urlencode patch
    build_and_apply_patch(
        pyramid_encode_module, "urlencode", urllib_patch.build_urlencode_patch
    )


def register_patches():
    register_post_import_hook(patch_session_pyramid, PYRAMID_SESSION_MODULE)
    register_post_import_hook(patch_encode_pyramid, "pyramid.encode")


def reverse_patches():
    pyramid_session = sys.modules.get(PYRAMID_SESSION_MODULE)
    pyramid_encode = sys.modules.get("pyramid.encode")

    if pyramid_session:
        patch_manager.reverse_patches_by_owner(pyramid_session)
    if pyramid_encode:
        patch_manager.reverse_patches_by_owner(pyramid_encode)
        urllib_patch.reverse_patches()



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import sys

from contrast.extern.wrapt import register_post_import_hook
from contrast.agent.policy import patch_manager
from contrast.utils.patch_utils import build_and_apply_patch


def build_noop_patch(*_):
    def noop_intern(string):
        """
        Defeat interning by returning the original string

        sys.intern takes no kwargs
        """
        return string

    return noop_intern


def patch_sys(module):
    build_and_apply_patch(module, "intern", build_noop_patch)


def register_patches():
    register_post_import_hook(patch_sys, "sys")


def reverse_patches():
    # sys is always imported, no need to check
    patch_manager.reverse_patches_by_owner(sys)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.extern import structlog as logging

logger = logging.getLogger("contrast")


class DisableReaction(object):
    NAME = "DISABLE"
    ENABLE = "enable"
    MESSAGE = "Contrast received instructions to disable itself - Disabling now"

    @staticmethod
    def run(settings):
        logger.warning(DisableReaction.MESSAGE)

        if settings.config:
            settings.config.put(DisableReaction.ENABLE, False)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.wsgi.middleware import WSGIMiddleware
from contrast.agent.middlewares.app_finder import get_original_app_or_fail
from contrast.utils.decorators import fail_quietly
from contrast.extern import structlog as logging
from contrast.agent.middlewares.route_coverage.bottle_routes import create_bottle_routes
from contrast.agent.middlewares.route_coverage.common import build_route
import bottle
from contrast.utils.decorators import cached_property

logger = logging.getLogger("contrast")


class BottleMiddleware(WSGIMiddleware):
    # Since Bottle is WSGI-based, there is no way to retrieve the app name.
    # Use common config to define an app name.
    def __init__(self, app, orig_bottle_app=None):
        self.bottle_app = (
            orig_bottle_app
            if orig_bottle_app is not None
            and isinstance(orig_bottle_app, bottle.Bottle)
            else get_original_app_or_fail(app, bottle.Bottle)
        )
        super().__init__(app, app_name="Bottle Application")

    @fail_quietly("Unable to get route coverage", return_value={})
    def get_route_coverage(self):
        return create_bottle_routes(self.bottle_app)

    @fail_quietly("Unable to get Bottle view func")
    def get_view_func(self, request):
        path = request.path
        if not path:
            return None
        method = request.method
        try:
            route_info = self.bottle_app.match(
                {"PATH_INFO": path, "REQUEST_METHOD": method}
            )
        except bottle.HTTPError:
            return None

        if not route_info:
            return None
        view_func = route_info[0].callback
        return view_func

    @fail_quietly("Unable to build route", return_value="")
    def build_route(self, view_func, url):
        return build_route(url, view_func)

    @cached_property
    def name(self):
        return "bottle"



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent import scope
from contrast.agent.assess.utils import is_tracked
from contrast.agent.assess.policy import string_propagation
from contrast.utils.decorators import fail_quietly


def contrast__cformat__modulo(format_str, args):
    with scope.propagation_scope():
        result = format_str % args

    if scope.in_scope() or scope.in_trigger_scope():
        return result

    _propagate_cformat(result, format_str, args)

    return result


@fail_quietly("failed to propagate through modulo in contrast__cformat__modulo")
def _propagate_cformat(result, format_str, args):
    propagation_func = None

    if isinstance(result, str):
        propagation_func = string_propagation.propagate_unicode_cformat
    elif isinstance(result, bytes):
        propagation_func = string_propagation.propagate_bytes_cformat
    elif isinstance(result, bytearray):
        # TODO: PYT-2709 - Update tests to verify bytearray is supported with and without funchook
        propagation_func = string_propagation.propagate_bytearray_cformat

    if propagation_func is None:
        return

    with scope.contrast_scope(), scope.propagation_scope():
        propagation_func(result, format_str, result, args, None)


def contrast__add(left, right):
    """
    This function replaces addition in the AST. We use double underscore to lower the
    probability of a name conflict.

    It is basically a pure-python translation of <type>_concat_new from str_concat.c.
    In the future, we should consider implementing as much of this in C as possible.
    """
    with scope.propagation_scope():
        result = left + right

    _propagate_add(left, right, result)

    return result


def contrast__append(target, value):
    orig_target = target
    with scope.propagation_scope():
        target += value

    _propagate_add(orig_target, value, target)

    return target


def contrast__fstring(*strings):
    # fstring formatting is equivalent to str.join
    return "".join(strings)


@fail_quietly("failed to propagate through addition in contrast__add")
def _propagate_add(left, right, result):
    propagation_func = None
    if isinstance(result, str):
        propagation_func = string_propagation.propagate_unicode_concat
    elif isinstance(result, bytes):
        propagation_func = string_propagation.propagate_bytes_concat
    elif isinstance(result, bytearray):
        propagation_func = string_propagation.propagate_bytearray_concat
    if propagation_func is None:
        return

    if len(result) < 2:
        return

    if scope.in_scope() or scope.in_trigger_scope():
        return

    if not (is_tracked(left) or is_tracked(right)):
        return

    with scope.contrast_scope(), scope.propagation_scope():
        propagation_func(result, left, result, (right,), {})


def populate_rewritten_functions(dest):
    for function in [
        contrast__add,
        contrast__append,
        contrast__fstring,
        contrast__cformat__modulo,
    ]:
        dest.setdefault(function.__name__, function)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from collections import OrderedDict, defaultdict
import json
from os import path

import contrast
from contrast.agent.policy.patch_location_policy import PatchLocationPolicy
from contrast.agent.assess.policy.deadzone_node import DeadZoneNode
from contrast.agent.assess.policy.propagation_node import PropagationNode
from contrast.agent.assess.policy.source_node import SourceNode
from contrast.agent.policy.trigger_node import TriggerNode
from contrast.agent.assess.rules.dataflow_rule import DataflowRule
from contrast.agent.assess.rules.non_dataflow_rule import NonDataflowRule
from contrast.agent.policy.constants import (
    JSON_NAME,
    PROPAGATION_KEY,
    SOURCES_KEY,
    TRIGGERS_KEY,
    DEADZONE_KEY,
    REWRITES_KEY,
)
from contrast.agent.assess.rules.providers.hardcoded_key import HardcodedKey
from contrast.agent.assess.rules.providers.hardcoded_password import HardcodedPassword
from contrast.agent.assess.rules.triggers import (
    HttpOnlyRule,
    SecureFlagMissingRule,
    SessionTimeoutRule,
    SessionRewritingRule,
)
from contrast.utils.singleton import Singleton
from contrast.utils.decorators import cached_property, fail_loudly

from contrast.extern import structlog as logging

logger = logging.getLogger("contrast")


POLICY_DATA_DIR = path.join(path.dirname(contrast.__file__), "data")


class Policy(Singleton):
    POLICY_JSON_PATH = path.join(POLICY_DATA_DIR, "policy.json")

    PROVIDER_CLASSES = [HardcodedKey, HardcodedPassword]
    TRIGGER_RULES = {
        "httponly": HttpOnlyRule,
        "session-timeout": SessionTimeoutRule,
        "secure-flag-missing": SecureFlagMissingRule,
        "session-rewriting": SessionRewritingRule,
    }

    def init(self):
        """
        Singletons should override init, not __init__.
        """
        self.sources = []
        self.propagators = []
        self.deadzones = []
        self.triggers = OrderedDict()
        self.providers = OrderedDict()
        self.rewrite_modules = []
        self.rewrite_classes = []

        self.propagators_by_name = defaultdict(list)
        # This is used to ensure that we only ever have one PatchLocationPolicy
        # instance per patch location.
        self.policy_by_name = OrderedDict()
        # This is used to represent a list of PatchLocationPolicy instances per each
        # module. It is what enables the application of module-level import hooks.
        self.policy_by_module = OrderedDict()

        self.protect_policy_by_module = OrderedDict()

        self.policy_data = {}

        self.initialize()

    def process_node(self, node):
        """
        Create patch location policy for node and update lookup tables if necessary
        """
        patch_policy = self.policy_by_name.setdefault(
            node.name, PatchLocationPolicy(node)
        )
        patch_policy.add_node(node)

        policy_list = self.policy_by_module.setdefault(node.module, [])
        policy_list.append(patch_policy)

        if isinstance(node, TriggerNode) and node.protect_mode:
            policy_list = self.protect_policy_by_module.setdefault(node.module, [])
            policy_list.append(patch_policy)

        return patch_policy

    def initialize(self):
        loaded_json = self._get_json_data()
        if not loaded_json:
            return

        self.policy_data = loaded_json

        self._load_sources()
        self._load_propagators()
        self._load_triggers()
        self._load_deadzones()
        self._load_rewrites()

        self._load_providers()

    def _load_rewrites(self):
        rewrites = self.policy_data[REWRITES_KEY]
        for item in rewrites["modules"]:
            self.rewrite_modules.append(item)
        for item in rewrites["classes"]:
            self.rewrite_classes.append(item)

    def _load_deadzones(self):
        for item in self.policy_data[DEADZONE_KEY]:
            dz = DeadZoneNode.from_dict(item)

            self.deadzones.append(dz)
            self.process_node(dz)

    def _load_sources(self):
        for framework, nodes in self.policy_data[SOURCES_KEY].items():
            for value in nodes:
                source_node = SourceNode.from_dict(framework, value)

                # Uncomment if we re-implement version_mismatch
                # if source_node.version_mismatch:
                #     continue

                if source_node.should_apply():
                    self.sources.append(source_node)
                    self.process_node(source_node)

    def _load_propagators(self):
        for item in self.policy_data[PROPAGATION_KEY]:
            prop = PropagationNode.from_dict(item)

            # Uncomment if we re-implement version_mismatch
            # if prop.version_mismatch:
            #     continue

            self.propagators.append(prop)
            self.propagators_by_name[prop.name].append(prop)
            self.process_node(prop)

    def _find_rule(self, json_node):
        """
        Find rule class corresponding to rule name

        First, look for special cases based on name. If none exist, determine whether
        to use dataflow or non-dataflow rule class based on JSON.
        """
        rule_name = json_node.get(JSON_NAME)
        trigger_rule = self.TRIGGER_RULES.get(rule_name)
        if trigger_rule:
            return trigger_rule

        is_dataflow = json_node.get(DataflowRule.JSON_DATAFLOW, True)
        return DataflowRule if is_dataflow else NonDataflowRule

    def _load_triggers(self):
        for item in self.policy_data[TRIGGERS_KEY]:
            rule_cls = self._find_rule(item)
            rule = rule_cls.from_dict(item)
            self.triggers[rule.name] = rule
            for node in rule.nodes:
                self.process_node(node)

    def _load_providers(self):
        for item in self.PROVIDER_CLASSES:
            instance = item()

            self.providers[instance.name] = instance

    @fail_loudly("Failed to read policy json file")
    def _get_json_data(self):
        with open(self.POLICY_JSON_PATH) as json_file:
            return json.load(json_file)

    @fail_loudly("Failed to add node to policy")
    def add_source_node(self, node):
        node.validate()
        self.sources.append(node)
        return self.process_node(node)

    @cached_property
    def string_method_nodes(self):
        return [
            node
            for node in self.policy_by_module["BUILTIN"]
            if node.class_name == "str"
        ]



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import logging
import os

from contrast.extern import structlog

from contrast.agent.rewriter import register
from contrast.agent.assess.policy.string_propagation import (
    build_string_propagator_functions,
)


REWRITE_FOR_PYTEST: str = "__CONTRAST_REWRITE_FOR_PYTEST"


def start_runner():
    if os.environ.get(REWRITE_FOR_PYTEST):
        structlog.configure(
            wrapper_class=structlog.make_filtering_bound_logger(logging.ERROR)
        )

        build_string_propagator_functions()
        register(override_settings=True)
        # Short-circuit initialization
        return

    # We need to avoid this import for the REWRITE_FOR_PYTEST case
    from contrast.agent.agent_state import initialize

    initialize(runner=True, name="agent runner")



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import gc
import inspect

from contrast.assess_extensions.cs_str import has_funchook
from contrast.agent import scope
from contrast.agent.assess.policy import string_propagation
from contrast.agent.policy import patch_manager
from contrast.utils.decorators import fail_quietly
from contrast.utils.patch_utils import build_and_apply_patch
from contrast.utils.namespace import Namespace


class module(Namespace):
    patched_classes = []


def find_str_subclasses():
    """
    Digs through the garbage (collector) to find any subclasses of str
    """
    for obj in gc.get_objects():
        try:
            if inspect.isclass(obj) and issubclass(obj, str):
                yield obj
        except Exception:
            continue


@fail_quietly("Failed to propagate string subclass __new__")
def propagate_cast(result, args, kwargs):
    with scope.contrast_scope():
        string_propagation.propagate_unicode_cast(
            result, args[0], result, args[1:], kwargs
        )


def build_cast_patch(orig_func, patch_policy):
    del patch_policy

    def __new__(*args, **kwargs):
        result = orig_func(*args, **kwargs)
        propagate_cast(result, args, kwargs)
        return result

    return __new__


@fail_quietly("Failed to apply patches for string subclasses")
def register_patches():
    """
    Applies patches to known subclasses of str

    Problem statement: the cast/new hook for subclasses of str does not work
    without funchook. This seems to be an issue with timing: subclasses that
    are defined *after* our (non-funchook) extension hooks are applied do not
    appear to have a problem.

    However, even with the runner, our instrumentation is not necessarily going
    to be early enough to affect *all* subclasses that are defined.

    The solution is to use the garbage collector to find any subclasses of str
    that may already be defined at this point in time. We then explicitly apply
    propagation patches to the __new__ methods of these classes.
    """
    if has_funchook():
        return

    for cls in find_str_subclasses():
        build_and_apply_patch(cls, "__new__", build_cast_patch)
        module.patched_classes.append(cls)


def reverse_patches():
    for cls in module.patched_classes:
        patch_manager.reverse_patches_by_owner(cls)

    module.patched_classes = []



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.rules.config import (
    FlaskHttpOnlyRule,
    FlaskSecureFlagRule,
    FlaskSessionAgeRule,
)
from contrast.agent.middlewares.route_coverage.common import (
    create_routes,
    build_route,
    get_view_func_for_request,
)
from contrast.wsgi.middleware import WSGIMiddleware
from contrast.utils.decorators import fail_quietly, cached_property


class FlaskMiddleware(WSGIMiddleware):
    def __init__(self, app):
        self.app = app

        self.config_rules = [
            FlaskSessionAgeRule(),
            FlaskSecureFlagRule(),
            FlaskHttpOnlyRule(),
        ]

        super().__init__(app.wsgi_app, app_name=app.name)

    @fail_quietly("Unable to get route coverage", return_value={})
    def get_route_coverage(self):
        return create_routes(self.app)

    @fail_quietly("Unable to get Flask view func")
    def get_view_func(self, request):
        return get_view_func_for_request(request, self.app)

    @fail_quietly("Unable to build route", return_value="")
    def build_route(self, view_func, url):
        return build_route(view_func.__name__, view_func)

    @fail_quietly("Failed to run config scanning rules")
    def _scan_configs(self):
        for rule in self.config_rules:
            rule.apply(self.app)

    @cached_property
    def name(self):
        return "flask"



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.extern.wrapt import register_post_import_hook

from contrast.agent.policy import patch_manager
from contrast.utils.patch_utils import build_and_apply_patch

DJANGO_WSGI_NAME = "django.core.wsgi"


def build_get_wsgi_app_patch(orig_func, patch_policy):
    del patch_policy

    def get_wsgi_application(*args, **kwargs):
        # Avoids circular import
        from contrast.django import ContrastMiddleware

        return ContrastMiddleware(orig_func(*args, **kwargs))

    return get_wsgi_application


def patch_django_wsgi(module):
    build_and_apply_patch(module, "get_wsgi_application", build_get_wsgi_app_patch)


def register_patches():
    register_post_import_hook(patch_django_wsgi, DJANGO_WSGI_NAME)


def reverse_patches():
    patch_manager.reverse_patches_by_owner(DJANGO_WSGI_NAME)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.api import Route
from contrast.utils.decorators import log_time
from contrast.agent.middlewares.route_coverage.common import build_key

from contrast.extern import structlog as logging

logger = logging.getLogger("contrast")


class RoutesMixin(object):
    """
    Class for route coverage work.
    Route coverage is assess only.
    """

    @log_time("route coverage")
    def handle_routes(self, context, request):
        """
        Method that should run for all middlewares immediately after the application executes.
        """
        if context and request:
            self.get_routes_on_first_request()

            self.check_for_new_routes(context, request)

            self.append_route_to_findings(context)

    def get_routes_on_first_request(self):
        """
        Attempt to do route discovery on the first request
        """
        if not self.first_request:
            return

        logger.debug("First request so looking for app routes.")

        routes = self.get_route_coverage()

        if not routes or not isinstance(routes, dict):
            logger.debug("Could not find any routes")
        else:
            self.routes = routes

            routes_to_log = [
                f"{route.verb}: {route.url}" for route in self.routes.values()
            ]
            logger.debug("Found the following routes: %s", routes_to_log)

    def check_for_new_routes(self, context, request):
        """
        We check for new routes (in addition to the ones already in self.routes which
        were found through looking for expected routes the app registered during the first request.

        This instead looks to see if the current request is a new route we should add
        to middleware routes.
        """
        logger.debug("Checking for new route.")

        func = self.get_view_func(request)
        if func is None:
            return

        request_method = request.method
        route_id = build_key(str(id(func)), request_method)

        self.update_route_information(context, route_id, request_method, func)

    def update_route_information(self, context, route_id, request_method, func):
        """
        Given a context and route_id, check if the route_id is in middleware (self) routes.
        Store the route as current and observed route for later use.

        :param context: RequestContext instance
        :param route_id: string id for a route
        :param request_method: string such as 'GET'
        :param func:  App view function
        :return: no return, side effects only
        """
        if route_id not in self.routes:
            self.extend_middleware_routes(
                request_method, context.request.get_normalized_uri(), route_id, func
            )

        route = self.routes[route_id]

        logger.debug("Route visited: %s : %s", route.verb, route.url)

        context.current_route = route

        # Currently we do not report an observed route if the route signature is empty.
        # As a team we've decided there isn't a meaningful default signature value
        # we can provide to customers. If a route doesn't show up in Contrast UI,
        # it may be due to its missing signature. In this scenario, we will have to work
        # with the customer directly to understand why the signature was not created.
        if not route.signature:
            logger.debug(
                "No route signature found for %s : %s (id=%s). Not updating observed"
                " route",
                route.verb,
                route.url,
                route_id,
            )
            return

        context.observed_route.signature = route.signature

        context.observed_route.url = route.url
        context.observed_route.verb = route.verb

    def extend_middleware_routes(self, request_method, url, route_id, func):
        route = self.build_route(func, url)
        self.routes[route_id] = Route(verb=request_method, url=url, route=route)

    def append_route_to_findings(self, context):
        """
        Route discovery and current route is not identified until the handle ensure
        part of the request lifecycle, after assess has analyzed and potentially created
        a finding, so that is why we have to append the now-available current route to
        the existing finding.
        """
        if not context.current_route:
            logger.debug("No current route to append to findings")
            return

        if not context.findings:
            logger.debug("No findings to append route to")
            return

        for finding in context.findings:
            if not finding.routes:
                logger.debug(
                    "Appending route %s:%s to %s",
                    context.current_route.verb,
                    context.current_route.url,
                    finding.rule_id,
                )
                finding.routes.append(context.current_route)

            finding.set_version()

    def get_view_func(self, request):
        return

    def build_route(self, view_func, url):
        return ""



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.extern import wrapt

import contrast
from contrast.agent.assess.policy.analysis import analyze
from contrast.agent.policy.loader import Policy
from contrast.utils.decorators import fail_quietly
from contrast.utils.patch_utils import build_and_apply_patch


STARLETTE_REQUESTS = "starlette.requests"
SESSION = "session"


class ContrastSessionDictProxy(wrapt.ObjectProxy):
    """
    Custom ObjectProxy we use to wrap dicts returned by starlette's request.session
    property. These proxied dicts have a trigger for trust-boundary-violation on
    __setitem__.
    """

    def __setitem__(self, key, value):
        result = None
        try:
            result = self.__wrapped__.__setitem__(key, value)
        finally:
            analyze_setitem(result, (self, key, value))

        return result


@fail_quietly("Failed to analyze session dict __setitem__")
def analyze_setitem(result, args):
    context = contrast.CS__CONTEXT_TRACKER.current()
    policy = Policy().policy_by_name.get("starlette.sessions.dict.__setitem__")
    analyze(context, policy, result, args, {})


def build_session_patch(orig_prop, patch_policy):
    def session_fget(*args, **kwargs):
        """
        Function used to replace fget for starlette's request.session property.
        This function returns proxied dictionaries - see ContrastSessionDictProxy.
        """
        session_dict = orig_prop.fget(*args, **kwargs)

        context = contrast.CS__CONTEXT_TRACKER.current()
        if context is None:
            return session_dict

        return ContrastSessionDictProxy(session_dict)

    return property(session_fget, orig_prop.fset, orig_prop.fdel)


def patch_starlette_session(requests_module):
    build_and_apply_patch(requests_module.Request, SESSION, build_session_patch)


def register_patches():
    wrapt.register_post_import_hook(patch_starlette_session, STARLETTE_REQUESTS)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from io import BytesIO

from contrast.agent import scope as scope_
from contrast.agent.assess.policy.analysis import skip_analysis
from contrast.agent.assess.policy.source_policy import (
    cs__apply_source,
    build_source_node,
)
from contrast.utils.decorators import fail_loudly, cached_property
from contrast.extern.webob.headers import ResponseHeaders

# note: request method is not a source, per labs
# source types translated from environ sources in environ_tracker.py
SCOPE_SOURCES = {
    "path": "URI",
    "raw_path": "URI",
    "query_string": "QUERYSTRING",
    "root_path": "OTHER",
    "server": "OTHER",
    "client": "URI",
    "headers": "HEADER",
    "cookie": "COOKIE",  # not an actual scope element; we use this for headers[cookie]
}

SOURCE_DICT = {"module": "ASGI.scope", "instance_method": False, "target": "RETURN"}
EVENT_TYPE = "http.request"


class ASGIRequest:
    """
    A barebones request class for ASGI applications. This class is mostly useful for
    managing the `receive` coroutine, which is how ASGI apps get the request body.
    """

    def __init__(self, scope, original_receive):
        self.scope = scope
        self.original_receive = original_receive

        self._called = False
        self._body = None
        self._wsgi_environ = None

    async def fake_receive(self):
        """
        This coroutine should be sent to the underlying ASGI application instead of the
        original `receive`. This allows us to capture the request body while still
        providing the ASGI app with a valid `receive`-like coroutine.
        """
        if self._called:
            return {"type": "http.disconnect"}

        body = await self.body()
        self._called = True
        return {
            "type": EVENT_TYPE,
            "body": body,
            "more_body": False,
        }

    async def body(self):
        """
        Call the original `receive` coroutine and store the result. This coroutine can
        be called multiple times without adverse consequences.
        """
        if self._body is not None:
            return self._body

        body_parts = []
        more_body = True
        while more_body:
            event = await self.original_receive()
            if event["type"] == EVENT_TYPE:
                body_parts.append(event.get("body", b""))
                more_body = event.get("more_body", False)
            else:
                more_body = False
        self._body = b"".join(body_parts)
        return self._body

    async def to_wsgi_environ(self):
        if self._wsgi_environ is not None:
            return self._wsgi_environ

        body = await self.body()
        self._wsgi_environ = _scope_to_environ(self.scope, body)
        return self._wsgi_environ


class ASGIResponse:
    """
    A Response class for ASGI applications. This class handles everything related to the
    `send` coroutine, which is how ASGI applications send response data.
    """

    def __init__(self, original_send):
        self._original_send = original_send
        self._response_started = False
        self._more_body = True
        self._status_code = None
        self._asgi_headers = []
        self._body_parts = []

    async def fake_send(self, event):
        """
        This is a heavily simplified version of the actual `send` implementation from
        Uvicorn.

        `send` is the function that the application uses to send data back to the client
        - by replacing it, we can analyze and control the response. Instead of actually
        sending data anywhere, this method simply stores the response content on our
        ASGIResponse object.
        """
        event_type = event["type"]
        if not self._response_started and self._more_body:
            if event_type != "http.response.start":
                raise RuntimeError(
                    f"Expected ASGI event http.response.start but got '{event_type}'"
                )
            self._response_started = True
            self._status_code = event["status"]
            self._asgi_headers = event.get("headers", [])
        elif self._more_body:
            if event_type != "http.response.body":
                raise RuntimeError(
                    f"Expected ASGI event http.response.body but got '{event_type}'"
                )
            self._body_parts.append(event.get("body", b""))
            self._more_body = event.get("more_body", False)
        else:
            raise RuntimeError(
                f"Unexpected ASGI event '{event_type}' after response already completed"
            )

    # TODO: PYT-2127 - it's possible that we may want to change these properties around
    # to make them writeable
    @cached_property
    def body(self):
        self._check_more_body()
        return b"".join(self._body_parts)

    @cached_property
    def headers(self):
        """
        To comply with our internal ResponseWrapper API, we need headers to be strings,
        not bytes
        """
        self._check_response_started()
        return ResponseHeaders(
            [(key.decode(), value.decode()) for key, value in self._asgi_headers]
        )

    @cached_property
    def status_code(self):
        self._check_response_started()
        return self._status_code

    async def call_send(self):
        """
        Call the original `send` method we captured from the server using the
        response information we captured from the application.
        """
        await self._original_send(
            {
                "type": "http.response.start",
                "status": self.status_code,
                "headers": self._asgi_headers,
            }
        )
        await self._original_send(
            {
                "type": "http.response.body",
                "body": self.body,
                "more_body": False,
            }
        )

    def _check_more_body(self):
        if self._more_body:
            raise RuntimeError(f"ASGI application did not finish sending response body")

    def _check_response_started(self):
        if not self._response_started:
            raise RuntimeError(f"ASGI application did not start response")


@fail_loudly("Failed to convert starlette request to environ dict")
async def starlette_request_to_environ(request):
    body = await _get_body_non_destructive(request)
    return _scope_to_environ(request.scope, body)


async def _get_body_non_destructive(request) -> bytes:
    """
    Get the request body without consuming the ASGI `receive` callable. This involves
    some trickery. We bank on the fact that we're in a starlette / fastAPI environment,
    which uses `call_next` for each successive middleware.

    In the general ASGI case, we will need to perform all of the work currently done in
    request.body(), then send our replaced_receive function to the next middleware or
    application when we call it.
    """
    body = await request.body()
    called = False

    async def replaced_receive():
        # On the first call, send the whole body
        # On successive calls, send the disconnect message
        nonlocal called
        if not called:
            called = True
            return {
                "type": EVENT_TYPE,
                "body": body,
                "more_body": False,
            }
        return {"type": "http.disconnect"}

    request._receive = replaced_receive
    return body


def _scope_to_environ(scope: dict, body: bytes) -> dict:
    """
    Convert an asgi `scope` into a wsgi `environ` dict

    Copied from https://github.com/django/asgiref/blob/main/asgiref/wsgi.py
    and modified to our needs
    """
    environ = {
        "REQUEST_METHOD": scope.get("method", ""),
        "SCRIPT_NAME": scope.get("root_path", "").encode("utf8").decode("latin1"),
        "PATH_INFO": scope.get("path", "").encode("utf8").decode("latin1"),
        "QUERY_STRING": scope["query_string"].decode("ascii"),
        "SERVER_PROTOCOL": f"HTTP/{scope.get('http_version', '')}",
        "wsgi.version": (1, 0),
        "wsgi.url_scheme": scope.get("scheme", "http"),
        "wsgi.input": BytesIO(body),
        "wsgi.errors": BytesIO(),
        "wsgi.multithread": True,
        "wsgi.multiprocess": True,
        "wsgi.run_once": False,
    }

    # Get server name and port - required in WSGI, not in ASGI
    if "server" in scope:
        environ["SERVER_NAME"] = scope["server"][0]
        environ["SERVER_PORT"] = str(scope["server"][1] or 80)
    else:
        environ["SERVER_NAME"] = "localhost"
        environ["SERVER_PORT"] = "80"

    if "client" in scope:
        environ["REMOTE_ADDR"] = scope["client"][0]

    # Go through headers and make them into environ entries
    for name, value in scope.get("headers", []):
        name = name.decode("latin1")
        if name == "content-length":
            corrected_name = "CONTENT_LENGTH"
        elif name == "content-type":
            corrected_name = "CONTENT_TYPE"
        else:
            corrected_name = f"HTTP_{name.upper().replace('-', '_')}"
        # HTTPbis say only ASCII chars are allowed in headers, but we latin1 just in
        # case
        value = value.decode("latin1")
        if corrected_name in environ:
            value = environ[corrected_name] + "," + value
        environ[corrected_name] = value

    return environ


def track_scope_sources(context, scope):
    """
    Iterate over the ASGI scope and explicitly track all relevant values. This is
    similar to the environ tracker. This function does not track the request body, which
    comes from the receive awaitable.

    There's an unfortunate possibility of starlette's request object having already
    cached some data from the scope dict. The easiest way around this would be to find
    and clear all of the places where starlette might save these variables. This could
    be avoided by going for an approach that's closer to pure ASGI instead of relying as
    much on starlette.

    @param context: the current request context
    @param scope: the ASGI scope dict
    """
    if skip_analysis(context):
        return

    with scope_.contrast_scope():
        for key, value in scope.items():
            if key in ["client", "server"]:
                for elem in value:
                    _track_scope_item(context, scope, key, elem)
            elif key == "headers":
                for header_key, header_value in value:
                    key = "cookie" if header_key == b"cookie" else "headers"
                    _track_scope_item(context, scope, key, header_key)
                    _track_scope_item(context, scope, key, header_value)
            else:
                _track_scope_item(context, scope, key, value)


def _track_scope_item(context, scope, key, value):
    # there are other elements in `scope`, but only those in SCOPE_SOURCES matter to us
    if key in SCOPE_SOURCES:
        node = build_source_node(SOURCE_DICT, key, SCOPE_SOURCES[key], "ASGI")
        cs__apply_source(context, node, value, scope, value, (), {}, source_name=key)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
"""
Rules for coverage:

- new routes on init get count 0
- new routes after init get count 1 per context
- routes without a method type specified on init are ['GET', 'POST']
- a route needs verb, route, url, and count
- url should be normalized uri
- route should be the path to the view function (aka controller for Java people)
- one method type per route

Example:
      GET /blog/foo/bar - app.blogs.find(request, ) 0
"""

import re

from contrast.api import Route

DEFAULT_ROUTE_METHODS = ("GET", "POST")

# 1 or more digits, then a group: either '/' or end of string
DIGITS_ONLY_PATH_COMPONENT = r"/\d+(/|$)"

# Retruns digits between '/' or end of string
FIND_REQEUST_PARAMS = r"(?<=\/)\d+(?=\/|$)"


def create_routes(app):
    """
    Returns all the routes registered to an app as a dict
    :param app: {Quart or Flask} app
    :return: dict {route_id:  api.Route}
    """
    routes = {}

    for rule in list(app.url_map.iter_rules()):
        view_func = app.view_functions[rule.endpoint]

        route = build_route(rule.endpoint, view_func)

        route_id = str(id(view_func))

        methods = rule.methods or DEFAULT_ROUTE_METHODS

        for method_type in methods:
            key = build_key(route_id, method_type)
            routes[key] = Route(
                verb=method_type,
                url=get_normalized_uri(str(rule)),
                route=route,
            )

    return routes


def get_normalized_uri(path):
    """
    A best-effort to remove client-specific information from the path.

    Example:
    /user/123456/page/12 -> /user/{n}/page/{n}
    """
    return re.sub(DIGITS_ONLY_PATH_COMPONENT, r"/{n}\1", path)


def get_url_parameters(path):
    """
    A best-effort to remove client-specific information from the path.

    Example:
    /user/123456/page/12 -> /user/{n}/page/{n}
    """
    return re.findall(FIND_REQEUST_PARAMS, path)


def build_route(view_func_name, view_func):
    view_func_args = build_args_from_function(view_func)
    return view_func_name + view_func_args


def get_view_func_for_request(request, app):
    """
    Find the view function for the current request in the app.
    :param request: current request
    :param app: {Quart or Flask} app
    :return: function
    """
    from werkzeug.exceptions import NotFound, MethodNotAllowed

    adapter = app.url_map.bind("empty")

    if None in (request, adapter):
        return None

    try:
        match = adapter.match(request.path, method=request.method)
    except (NotFound, MethodNotAllowed):
        match = None

    func = None
    if match is not None:
        func = app.view_functions[match[0]]

    return func


def build_key(route_id, method):
    return "_".join([route_id, method])


def build_args_from_function(func):
    """
    Attempts to grab argument names from the function definition.

    Defaults to () if none exist
    If there is no view function, like in the case of a pure WSGI app, then the func will
    be a string like '/sqli' and we just return that.

    """
    method_arg_names = "()"
    if func is not None and hasattr(func, "__code__"):
        method_arg_names = str(
            func.__code__.co_varnames[0 : func.__code__.co_argcount]
        ).replace("'", "")
    elif isinstance(func, str):
        method_arg_names = func

    return method_arg_names



# SPDX-License-Identifier: MIT OR Apache-2.0
# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the MIT License.  See the LICENSE file in the root of this
# repository for complete details.


from __future__ import annotations

from contrast.extern.structlog import (
    contextvars,
    dev,
    processors,
    stdlib,
    testing,
    threadlocal,
    tracebacks,
    types,
    typing,
)
from contrast.extern.structlog._base import BoundLoggerBase, get_context
from contrast.extern.structlog._config import (
    configure,
    configure_once,
    get_config,
    get_logger,
    getLogger,
    is_configured,
    reset_defaults,
    wrap_logger,
)
from contrast.extern.structlog._generic import BoundLogger
from contrast.extern.structlog._log_levels import make_filtering_bound_logger
from contrast.extern.structlog._output import (
    BytesLogger,
    BytesLoggerFactory,
    PrintLogger,
    PrintLoggerFactory,
    WriteLogger,
    WriteLoggerFactory,
)
from contrast.extern.structlog.exceptions import DropEvent
from contrast.extern.structlog.testing import ReturnLogger, ReturnLoggerFactory


try:
    from contrast.extern.structlog import twisted
except ImportError:
    twisted = None  # type: ignore[assignment]


__title__ = "structlog"

__author__ = "Hynek Schlawack"

__license__ = "MIT or Apache License, Version 2.0"
__copyright__ = "Copyright (c) 2013 " + __author__


__all__ = [
    "BoundLogger",
    "BoundLoggerBase",
    "BytesLogger",
    "BytesLoggerFactory",
    "configure_once",
    "configure",
    "contextvars",
    "dev",
    "DropEvent",
    "get_config",
    "get_context",
    "get_logger",
    "getLogger",
    "is_configured",
    "make_filtering_bound_logger",
    "PrintLogger",
    "PrintLoggerFactory",
    "processors",
    "reset_defaults",
    "ReturnLogger",
    "ReturnLoggerFactory",
    "stdlib",
    "testing",
    "threadlocal",
    "tracebacks",
    "twisted",
    "types",
    "typing",
    "wrap_logger",
    "WriteLogger",
    "WriteLoggerFactory",
]


def __getattr__(name: str) -> str:
    dunder_to_metadata = {
        "__version__": "version",
        "__description__": "summary",
        "__uri__": "",
        "__email__": "",
    }
    if name not in dunder_to_metadata.keys():
        raise AttributeError(f"module {__name__} has no attribute {name}")

    import sys
    import warnings

    if sys.version_info < (3, 8):
        from importlib_metadata import metadata
    else:
        from importlib.metadata import metadata

    warnings.warn(
        f"Accessing structlog.{name} is deprecated and will be "
        "removed in a future release. Use importlib.metadata directly "
        "to query for structlog's packaging metadata.",
        DeprecationWarning,
        stacklevel=2,
    )

    meta = metadata("structlog")

    if name == "__uri__":
        return meta["Project-URL"].split(" ", 1)[-1]
    elif name == "__email__":
        return meta["Author-email"].split("<", 1)[1].rstrip(">")

    return meta[dunder_to_metadata[name]]



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import tempfile
import time
from glob import glob
from io import open
from os import environ, path, system
from shutil import rmtree
from sys import platform, version_info
from setuptools import Extension, find_packages, setup
from setuptools.command.build_ext import build_ext
from setuptools.command.install_lib import install_lib

root_dir = path.abspath(path.dirname(__file__))


# This is where the version should be updated for releases
CONTRAST_RELEASE_VERSION = "5.19.2"

AGENT_LIB_REQUIREMENT = "~=0.5.4"


def read(*parts):
    with open(path.join(root_dir, *parts), encoding="utf-8") as f:
        return f.read()


extension_path = path.join("contrast", "assess_extensions")
extension_source_dir = path.join("src", extension_path)
version_dir = path.join(extension_source_dir, "py{}{}".format(*version_info[:2]))
common_dir = path.join(extension_source_dir, "common")

if platform.startswith("darwin"):
    link_args = ["-rpath", "@loader_path"]
    platform_args = []
else:
    platform_args = ["-Wno-cast-function-type"]
    link_args = []

debug = environ.get("ASSESS_DEBUG")
debug_args = ["-g", "-O1"] if debug else []
macros = [("ASSESS_DEBUG", 1)] if debug else []
macros.append(("EXTENSION_BUILD_TIME", '"{}"'.format(time.ctime())))

strict_build_args = ["-Werror"] if environ.get("CONTRAST_STRICT_BUILD") else []

NO_FUNCHOOK = environ.get("CONTRAST_NO_FUNCHOOK")
if NO_FUNCHOOK:
    macros.append(("NO_FUNCHOOK", 1))
    c_sources = [
        path.join(common_dir, name)
        for name in [
            "patches.c",
            "scope.c",
            "logging.c",
            "intern.c",
            "propagate.c",
            "format.c",
            "repr.c",
            "repeat.c",
            "streams.c",
            "subscript.c",
            "cast.c",
        ]
    ]
    c_sources.extend(glob(path.join(extension_source_dir, "py3", "patches.c")))
    c_sources.extend(
        [
            path.join(version_dir, name + ".c")
            for name in ["iobase", "bytesio", "stringio"]
        ]
    )
    libraries = []
else:
    c_sources = glob(path.join(common_dir, "*.c")) + glob(path.join(version_dir, "*.c"))
    # Add source files common to all python3 versions
    c_sources.extend(glob(path.join(extension_source_dir, "py3", "*.c")))
    libraries = ["funchook"]


extensions = [
    Extension(
        "contrast.assess_extensions.cs_str",
        c_sources,
        libraries=libraries,
        include_dirs=[
            extension_source_dir,
            path.join(extension_source_dir, "include"),
        ],
        library_dirs=[extension_source_dir],
        # Path relative to the .so itself (works for gnu-ld)
        runtime_library_dirs=["$ORIGIN"],
        extra_compile_args=[
            "-Wall",
            "-Wextra",
            "-Wno-unused-parameter",
            "-Wmissing-field-initializers",
        ]
        + strict_build_args
        + debug_args
        + platform_args,
        extra_link_args=link_args,
        define_macros=macros,
    )
]

tempdir = None
funchook_temp = None


build_err_msg = """Failed to build Contrast C extension.
It is necessary for autotools (autoconf, automake) to be installed in order for
Contrast to build properly. On lightweight systems such as Alpine, it may be
necessary to install linux-headers if they are not available already. Some
other systems may require "build essential" packages to be installed.
"""


class ContrastBuildExt(build_ext):
    def run(self):
        if NO_FUNCHOOK:
            build_ext.run(self)
            return

        if system("/bin/sh src/contrast/assess_extensions/build_funchook.sh") != 0:
            raise RuntimeError(build_err_msg)

        build_ext.run(self)

        global tempdir
        global funchook_temp

        ext = "dylib" if platform.startswith("darwin") else "so"
        funchook_name = "libfunchook.{}".format(ext)
        funchook = path.join(extension_source_dir, funchook_name)

        tempdir = tempfile.mkdtemp("contrast-build")
        funchook_temp = path.join(tempdir, funchook_name)
        self.copy_file(funchook, funchook_temp)


class ContrastInstallLib(install_lib):
    def run(self):
        install_lib.run(self)

        if NO_FUNCHOOK:
            return

        if funchook_temp is not None:
            dest_dir = path.join(self.install_dir, extension_path)
            self.copy_file(funchook_temp, dest_dir)
            rmtree(tempdir)


if int(environ.get("CONTRAST_DEV_VERSION", "0")):
    version_config = dict(
        use_scm_version={
            # Parse out the issue name as prefix if it exists
            "tag_regex": (
                r"^(?P<prefix>(CONTRAST|PYT)-\d+/)?(?P<version>.*)(?P<suffix>)?"
            ),
            "git_describe_command": "git describe --tags --dirty --long",
        },
        setup_requires=["setuptools_scm"],
    )
else:
    version_config = dict(version=CONTRAST_RELEASE_VERSION)


setup(
    name="contrast-agent",
    description="Contrast Security's agent for Python web frameworks",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    # The project's main homepage.
    url="https://www.contrastsecurity.com",
    project_urls={
        "Change Log": "https://docs.contrastsecurity.com/en/python-agent-release-notes-and-archive.html",
        "Support": "https://support.contrastsecurity.com",
        "Trouble Shooting": "https://support.contrastsecurity.com/hc/en-us/search?utf8=%E2%9C%93&query=Python",
        "Wiki": "https://docs.contrastsecurity.com/",
    },
    # Author details
    author="Contrast Security, Inc.",
    author_email="python@contrastsecurity.com",
    # Choose your license
    license="CONTRAST SECURITY (see LICENSE.txt)",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        # Audience
        "Intended Audience :: Developers",
        # License; commercial
        # supported languages
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    keywords="security development",
    # See MANIFEST.in for excluded packages
    packages=find_packages("src"),
    package_dir={"": "src"},
    package_data={"contrast": ["../data/policy.json"]},
    python_requires=">=3.7,<3.12",
    # NOTE: if install_requirements changes, update license_info.py accordingly
    install_requires=[
        "requests>=2.4.2,<3",
        f"contrast-agent-lib{AGENT_LIB_REQUIREMENT}",
    ],
    include_package_data=True,
    cmdclass=dict(build_ext=ContrastBuildExt, install_lib=ContrastInstallLib),
    ext_modules=extensions,
    extras_require={},
    entry_points={
        "console_scripts": [
            "contrast-python-run = contrast.scripts:runner",
            "contrast-fix-interpreter-permissions = contrast.scripts:fix_interpreter_permissions",
            "contrast-validate-config = contrast.scripts:validate_config",
            "contrast-propagator-check = contrast.scripts:propagator_check",
        ]
    },
    **version_config,
)
